"""
The module ``learner`` provides the ability to run the segmentation algorithm.

===============================================================================
Overview
===============================================================================
The ``main`` function actually starts the algorithm and can
be called externally. The module allows for running multiple jobs through
|subprocess|_ or |drmaa|_. Configuration files for the learner are
provided in the examples_ and are entitled learner.

.. |subprocess| replace:: ``subprocess``
.. _subprocess: http://docs.python.org/2/library/subprocess.html
.. |drmaa| replace:: ``drmaa``
.. _drmaa: http://github.com/pygridtools/drmaa-python
.. _examples: http://github.com/nanshe-org/nanshe/tree/master/examples

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Apr 09, 2014 16:00:40 EDT$"


import os
import json
import itertools
import multiprocessing
import subprocess
import time

# Generally useful and fast to import so done immediately.
import numpy

import h5py

import nanshe

# Need in order to have logging information no matter what.
from nanshe.util import prof

from nanshe.util import iters, xnumpy,\
    wrappers, pathHelpers

from nanshe.io import hdf5

# Short function to process image data.
from nanshe.imp import segment

# For IO. Right now, just includes read_parameters for reading a config file.
from nanshe.io import xjson


# Get the loggers
trace_logger = prof.getTraceLogger(__name__)
logger = prof.logging.getLogger(__name__)



@prof.log_call(trace_logger)
def generate_neurons_io_handler(input_filename,
                                output_filename,
                                parameters_filename):
    """
        Uses generate_neurons to process a input_filename (HDF5 dataset) and
        outputs results to an output_filename (HDF5 dataset). Also,

        Args:
            input_filename          HDF5 filename to read from (should be a
                                    path to a h5py.Dataset)

            output_filename         HDF5 filename to write to (should be a path
                                    to a h5py.Group)

            parameters_filename     JSON filename with parameters.
    """

    # Extract and validate file extensions.

    # Parse parameter filename and validate that the name is acceptable
    parameters_filename_details = pathHelpers.PathComponents(
        parameters_filename
    )
    parameters_filename_ext = parameters_filename_details.extension
    parameters_filename_ext = parameters_filename_ext.lower().lstrip(os.extsep)
    # Clean up the extension so it fits the standard.
    if (parameters_filename_ext not in ["json"]):
        raise Exception(
            "Parameter file with filename: \"" + parameters_filename + "\"" +
            " provided with an unknown file extension: \"" +
            parameters_filename_ext + "\". If it is a " +
            "supported format, please run the given file through " +
            "nanshe_converter first before proceeding."
        )

    # Parse the parameters from the json file.
    parameters = xjson.read_parameters(parameters_filename)

    if (len(parameters) == 1) and ("generate_neurons_blocks" in parameters):
        generate_neurons_blocks(
            input_filename,
            output_filename,
            **parameters["generate_neurons_blocks"]
        )
    else:
        generate_neurons_a_block(input_filename, output_filename, **parameters)


@prof.log_call(trace_logger)
def generate_neurons_a_block(input_filename,
                             output_filename,
                             debug=False,
                             **parameters):
    """
        Uses generate_neurons to process a input_filename (HDF5 dataset) and
        outputs results to an output_filename (HDF5 dataset).

        Args:
            input_filename          HDF5 filename to read from (should be a
                                    path to a h5py.Dataset)

            output_filename         HDF5 filename to write to (should be a path
                                    to a h5py.Group)

            parameters              how the run should be configured.
    """

    # Extract and validate file extensions.

    # Parse input filename and validate that the name is acceptable
    input_filename_details = pathHelpers.PathComponents(input_filename)
    # Clean up the extension so it fits the standard.
    input_filename_ext = input_filename_details.extension
    input_filename_ext = input_filename_ext.lower().lstrip(os.extsep)
    if (input_filename_ext not in ["h5", "hdf5", "he5"]):
        raise Exception(
            "Input file with filename: \"" + input_filename + "\"" +
            " provided with an unknown file extension: \"" +
            input_filename_ext + "\". If it is a supported " +
            "format, please run the given file through " +
            "nanshe_converter first before proceeding."
        )

    # Parse output filename and validate that the name is acceptable
    output_filename_details = pathHelpers.PathComponents(output_filename)
    # Clean up the extension so it fits the standard.
    output_filename_ext = output_filename_details.extension
    output_filename_ext = output_filename_ext.lower().lstrip(os.extsep)
    if (output_filename_ext not in ["h5", "hdf5", "he5"]):
        raise Exception(
            "Output file with filename: \"" + output_filename + "\"" +
            " provided with an unknown file extension: \"" +
            output_filename_ext + "\". If it is a supported " +
            "format, please run the given file through nanshe_converter " +
            "first before proceeding."
        )

    # Where the original images are.
    input_dataset_name = input_filename_details.internalPath

    # Name of the group where all data will be stored.
    output_group_name = output_filename_details.internalPath


    # Read the input data.
    original_images = None
    with h5py.File(input_filename_details.externalPath, "r") as input_file_handle:
        original_images = hdf5.serializers.read_numpy_structured_array_from_HDF5(
            input_file_handle, input_dataset_name)
        original_images = original_images.astype(numpy.float32)

    # Write out the output.
    with h5py.File(output_filename_details.externalPath, "a") as output_file_handle:
        # Create a new output directory if doesn't exists.
        output_file_handle.require_group(output_group_name)

        # Group where all data will be stored.
        output_group = output_file_handle[output_group_name]

        # Create a soft link to the original images. But use the appropriate type of soft link depending on whether
        # the input and output file are the same.
        if "original_images" not in output_group:
            if input_filename_details.externalPath == output_filename_details.externalPath:
                output_group["original_images"] = h5py.SoftLink(
                    input_dataset_name
                )
            else:
                output_group["original_images"] = h5py.ExternalLink(
                    input_filename_details.externalPath, input_dataset_name
                )

        # Get a debug logger for the HDF5 file (if needed)
        array_debug_recorder = hdf5.record.generate_HDF5_array_recorder(
            output_group,
            group_name="debug",
            enable=debug,
            overwrite_group=False,
            recorder_constructor=hdf5.record.HDF5EnumeratedArrayRecorder
        )

        # Saves intermediate result to make resuming easier
        resume_logger = hdf5.record.generate_HDF5_array_recorder(
            output_group,
            recorder_constructor=hdf5.record.HDF5ArrayRecorder,
            overwrite=True
        )

        # Generate the neurons and attempt to resume if possible
        generate_neurons.resume_logger = resume_logger
        generate_neurons.recorders.array_debug_recorder = array_debug_recorder
        generate_neurons(
            original_images=original_images, **parameters["generate_neurons"]
        )

        # Save the configuration parameters in the attributes as a string.
        if "parameters" not in output_group.attrs:
            # Write the configuration parameters in the attributes as a string.
            output_group.attrs["parameters"] = repr(
                dict(list(parameters.items()) + [("debug", debug)])
            )


@prof.log_call(trace_logger)
def generate_neurons_blocks(input_filename,
                            output_filename,
                            num_processes=multiprocessing.cpu_count(),
                            block_shape=None,
                            num_blocks=None,
                            half_window_shape=None,
                            half_border_shape=None,
                            use_drmaa=False,
                            num_drmaa_cores=16,
                            debug=False,
                            **parameters):
    # TODO: Move function into new module with its own command line interface.
    # TODO: Heavy refactoring required on this function.

    # Extract and validate file extensions.

    # Parse input filename and validate that the name is acceptable
    input_filename_details = pathHelpers.PathComponents(input_filename)
    # Clean up the extension so it fits the standard.
    input_filename_ext = input_filename_details.extension
    input_filename_ext = input_filename_ext.lower().lstrip(os.extsep)
    if (input_filename_ext not in ["h5", "hdf5", "he5"]):
        raise Exception(
            "Input file with filename: \"" + input_filename + "\"" +
            " provided with an unknown file extension: \"" +
            input_filename_ext + "\". If it is a supported " +
            "format, please run the given file through " +
            "nanshe_converter first before proceeding."
        )

    # Parse output filename and validate that the name is acceptable
    output_filename_details = pathHelpers.PathComponents(output_filename)
    # Clean up the extension so it fits the standard.
    output_filename_ext = output_filename_details.extension
    output_filename_ext = output_filename_ext.lower().lstrip(os.extsep)
    if (output_filename_ext not in ["h5", "hdf5", "he5"]):
        raise Exception(
            "Output file with filename: \"" + output_filename + "\"" +
            " provided with an unknown file extension: \"" +
            output_filename_ext + "\". If it is a supported " +
            "format, please run the given file through nanshe_converter " +
            "first before proceeding."
        )

    # Where the original images are.
    input_dataset_name = input_filename_details.internalPath

    # Name of the group where all data will be stored.
    output_group_name = output_filename_details.internalPath

    # Directory where individual block runs will be stored.
    intermediate_output_dir = output_filename_details.externalPath.rsplit(
        output_filename_details.extension, 1)[0] + "_blocks"


    # Read the input data.
    original_images_shape_array = None
    with h5py.File(input_filename_details.externalPath, "r") as input_file_handle:
        original_images_shape_array = numpy.array(
            input_file_handle[input_dataset_name].shape
        )

    # Get the amount of the border to slice
    half_border_shape_array = None
    if half_border_shape is None:
        half_border_shape_array = numpy.zeros(
            len(original_images_shape_array), dtype=int
        )
    else:
        assert (len(half_window_shape) == len(original_images_shape_array))

        half_border_shape_array = numpy.array(half_border_shape)

        # Should be of type integer
        assert (issubclass(half_border_shape_array.dtype.type, numpy.integer))

        # Should not cut along temporal portion.
        # Maybe replace with a warning.
        assert (half_border_shape[0] == 0)

    # TODO: Refactor to expanded_numpy.
    # Cuts boundaries from original_images_shape
    original_images_pared_shape_array = original_images_shape_array - \
                                        2*half_border_shape_array

    # At least one of them must be specified. If not some mixture of both.
    assert ((block_shape is not None) or (num_blocks is not None))

    # Size of the block to use by pixels
    block_shape_array = None
    block_shape_array_undefined = None
    if block_shape is None:
        block_shape_array = -numpy.ones(
            original_images_pared_shape_array.shape, dtype=int
        )
        block_shape_array_undefined = numpy.ones(
            original_images_pared_shape_array.shape, dtype=bool
        )
    else:
        # Should have the same number of values in each
        assert (len(original_images_pared_shape_array) == len(block_shape))

        block_shape_array = numpy.array(block_shape, dtype=int)

        # Should be of type integer
        assert issubclass(block_shape_array.dtype.type, numpy.integer)

        block_shape_array_undefined = (block_shape_array == -1)

    # Number of
    num_blocks_array = None
    num_blocks_array_undefined = None
    if num_blocks is None:
        num_blocks_array = - \
            numpy.ones(original_images_pared_shape_array.shape, dtype=int)
        num_blocks_array_undefined = numpy.ones(
            original_images_pared_shape_array.shape, dtype=bool)
    else:
        # Should have the same number of values in each
        assert (len(original_images_pared_shape_array) == len(num_blocks))

        num_blocks_array = numpy.array(num_blocks, dtype=int)

        # Should be of type integer
        assert issubclass(num_blocks_array.dtype.type, numpy.integer)

        num_blocks_array_undefined = (num_blocks_array == -1)

    # Want to ensure that both aren't defined.
    assert ~(~block_shape_array_undefined & ~num_blocks_array_undefined).all()

    # If both are undefined, then the block should span that dimension
    missing_both = (block_shape_array_undefined & num_blocks_array_undefined)
    block_shape_array[
        missing_both] = original_images_pared_shape_array[missing_both]
    num_blocks_array[missing_both] = 1
    # Thus, we have resolved these values and can continue.
    block_shape_array_undefined[missing_both] = False
    num_blocks_array_undefined[missing_both] = False

    # Replace undefined values in block_shape_array
    missing_block_shape_array, block_shape_array_remainder = divmod(
        original_images_pared_shape_array[block_shape_array_undefined],
        num_blocks_array[block_shape_array_undefined]
    )
    # Block shape must be well defined.
    assert (block_shape_array_remainder == 0).all()
    missing_block_shape_array = missing_block_shape_array.astype(int)
    block_shape_array[block_shape_array_undefined] = missing_block_shape_array

    # Replace undefined values in num_blocks_array
    missing_num_blocks_array, num_blocks_array_remainder = divmod(
        original_images_pared_shape_array[num_blocks_array_undefined],
        block_shape_array[num_blocks_array_undefined]
    )
    # Allow some blocks to be smaller
    missing_num_blocks_array += (num_blocks_array_remainder != 0).astype(int)
    num_blocks_array[num_blocks_array_undefined] = missing_num_blocks_array
    # Get the overlap window
    half_window_shape_array = None
    if half_window_shape is None:
        half_window_shape_array = block_shape_array / 2.0
    else:
        assert (len(half_window_shape) == len(
            original_images_pared_shape_array))

        half_window_shape_array = numpy.array(half_window_shape)

        assert issubclass(half_window_shape_array.dtype.type, numpy.integer)

    # Want to make our window size is at least as large as the one used for
    # the f0 calculation.
    if "extract_f0" in parameters["generate_neurons"]["preprocess_data"]:
        #assert (parameters["generate_neurons"]["preprocess_data"]["extract_f0"]["half_window_size"] == half_window_shape_array[0])
        assert (parameters["generate_neurons"]["preprocess_data"]["extract_f0"]["half_window_size"] <= half_window_shape_array[0])

    # Estimate bounds for each slice. Uses typical python [begin, end) for the
    # indices.
    estimated_bounds = numpy.zeros(
        tuple(num_blocks_array),
        dtype=(int, original_images_pared_shape_array.shape + (2,))
    )

    for each_block_indices in iters.index_generator(*num_blocks_array):
        for each_dim, each_block_dim_index in enumerate(each_block_indices):
            estimated_lower_bound = each_block_dim_index * block_shape_array[each_dim]
            estimated_upper_bound = (each_block_dim_index + 1) * block_shape_array[each_dim]

            estimated_bounds[each_block_indices][each_dim] = numpy.array([
                estimated_lower_bound, estimated_upper_bound
            ])

    original_images_pared_slices = numpy.zeros(
        estimated_bounds.shape[:-2],
        dtype=[("actual", int, estimated_bounds.shape[-2:]),
               ("windowed", int, estimated_bounds.shape[-2:]),
               ("windowed_stack_selection", int, estimated_bounds.shape[-2:]),
               ("windowed_block_selection", int, estimated_bounds.shape[-2:])])

    # Get the slice that is within bounds
    original_images_pared_slices["actual"] = estimated_bounds
    original_images_pared_slices["actual"][..., 0] = numpy.where(
        0 < original_images_pared_slices["actual"][..., 0],
        original_images_pared_slices["actual"][..., 0],
        0
    )
    original_images_pared_slices["actual"][..., 1] = numpy.where(
        original_images_pared_slices["actual"][..., 1] < original_images_pared_shape_array,
        original_images_pared_slices["actual"][..., 1],
        original_images_pared_shape_array
    )

    # Gets the defined half_window_size.
    window_addition = numpy.zeros(estimated_bounds.shape, dtype=int)
    window_addition[..., 0] = -half_window_shape_array
    window_addition[..., 1] = half_window_shape_array

    # Get the slice with a window added.
    original_images_pared_slices[
        "windowed"] = estimated_bounds + window_addition
    original_images_pared_slices["windowed"][..., 0] = numpy.where(
        0 < original_images_pared_slices["windowed"][..., 0],
        original_images_pared_slices["windowed"][..., 0],
        0
    )
    original_images_pared_slices["windowed"][..., 1] = numpy.where(
        original_images_pared_slices["windowed"][..., 1] < original_images_pared_shape_array,
        original_images_pared_slices["windowed"][..., 1],
        original_images_pared_shape_array
    )

    # Get the slice information to get the windowed block from the original
    # image stack.
    original_images_pared_slices["windowed_stack_selection"] = original_images_pared_slices["windowed"]
    original_images_pared_slices["windowed_stack_selection"] += xnumpy.expand_view(
        half_border_shape_array, reps_after=2
    )

    # Get slice information for the portion within
    # `original_images_pared_slices["windowed"]`, which corresponds to
    # `original_images_pared_slices["actual"]`.
    #original_images_pared_slices["windowed_block_selection"][..., 0] = 0
    original_images_pared_slices["windowed_block_selection"][..., 1] = (
        original_images_pared_slices["actual"][..., 1] - original_images_pared_slices["actual"][..., 0]
    )
    original_images_pared_slices["windowed_block_selection"][:] += xnumpy.expand_view(
        original_images_pared_slices["actual"][..., 0] - original_images_pared_slices["windowed"][..., 0],
        reps_after=2
    )

    # Get a directory for intermediate results.
    try:
        os.mkdir(intermediate_output_dir)
    except OSError:
        # If it already exists, that is fine.
        pass

    intermediate_config = intermediate_output_dir + "/" + "config.json"

    # Overwrite the config file always
    with open(intermediate_config, "w") as fid:
        json.dump(
            dict(list(parameters.items()) + list({"debug" : debug}.items())),
            fid,
            indent=4,
            separators=(",", " : ")
        )
        fid.write("\n")

    # Construct an HDF5 file for each block
    input_filename_block = []
    output_filename_block = []
    stdout_filename_block = []
    stderr_filename_block = []
    with h5py.File(output_filename_details.externalPath, "a") as output_file_handle:
        # Create a new output directory if doesn't exists.
        output_file_handle.require_group(output_group_name)

        output_group = output_file_handle[output_group_name]

        if "original_images" not in output_group:
            if input_filename_details.externalPath == output_filename_details.externalPath:
                output_group["original_images"] = h5py.SoftLink(
                    input_dataset_name
                )
            else:
                output_group["original_images"] = h5py.ExternalLink(
                    input_filename_details.externalPath,
                    "/" + input_dataset_name
                )

        output_group.require_group("blocks")

        output_group_blocks = output_group["blocks"]

        input_file_handle = None
        try:
            # Skipping using region refs.
            input_file_handle = h5py.File(
                input_filename_details.externalPath, "r"
            )
        except IOError:
            # File is already open
            input_file_handle = output_file_handle

        for i, i_str, sequential_block_i in iters.filled_stringify_enumerate(
                original_images_pared_slices.flat
        ):
            intermediate_basename_i = intermediate_output_dir + "/" + i_str

            # Hold redirected stdout and stderr for each subprocess.
            stdout_filename_block.append(
                intermediate_basename_i + os.extsep + "out")
            stderr_filename_block.append(
                intermediate_basename_i + os.extsep + "err")

            # Ensure that the blocks are corrected to deal with trimming of the image stack
            # Must be done after the calculation of
            # original_images_pared_slices["windowed_block_selection"].
            sequential_block_i_windowed = sequential_block_i["windowed_stack_selection"]
            slice_i = tuple(
                slice(_1, _2, 1) for _1, _2 in sequential_block_i_windowed
            )

            if i_str not in output_group_blocks:
                output_group_blocks[i_str] = []
                output_group_blocks[i_str].attrs["filename"] = input_file_handle.filename
                output_group_blocks[i_str].attrs["dataset"] = input_dataset_name
                output_group_blocks[i_str].attrs["slice"] = str(slice_i)

            block_i = output_group_blocks[i_str]

            with h5py.File(intermediate_basename_i + os.extsep + "h5", "a") as each_block_file_handle:
                # Create a soft link to the original images. But use the
                # appropriate type of soft link depending on whether
                # the input and output file are the same.
                if "original_images" not in each_block_file_handle:
                    each_block_file_handle["original_images"] = h5py.ExternalLink(
                        os.path.relpath(
                            block_i.file.filename, intermediate_output_dir
                        ),
                        block_i.name
                    )

                input_filename_block.append(
                    each_block_file_handle.filename + "/" + "original_images"
                )
                output_filename_block.append(
                    each_block_file_handle.filename + "/"
                )

        if input_file_handle != output_file_handle:
            input_file_handle.close()

    cur_module_dirpath = os.path.dirname(os.path.dirname(nanshe.__file__))
    cur_module_filepath = os.path.splitext(os.path.abspath(__file__))[0]
    cur_module_name = os.path.relpath(cur_module_filepath, cur_module_dirpath)
    cur_module_name = cur_module_name.replace(os.path.sep, ".")
    cur_module_filepath += os.extsep + "py"

    import sys

    python = sys.executable

    executable_run = ""
    executable_run += "from sys import argv, path, exit; "

    executable_run += "path[:] = [\"%s\"] + [_ for _ in path if _ != \"%s\"]; " % \
                      (cur_module_dirpath, cur_module_dirpath,)
    executable_run += "from %s import main; exit(main(*argv))" % \
                      (cur_module_name,)

    block_process_args_gen = iters.izip(
        itertools.repeat(python),
        itertools.repeat("-c"),
        itertools.repeat(executable_run),
        itertools.repeat(intermediate_config),
        input_filename_block,
        output_filename_block,
        stdout_filename_block,
        stderr_filename_block
    )

    if use_drmaa:
        # Attempt to import drmaa.
        # If it fails to import, either the user has no intent in using it or
        # forgot to install it. If it imports, but fails to find symbols,
        # then the user has not set DRMAA_LIBRARY_PATH or
        # does not have libdrmaa.so.
        try:
            import drmaa
        except ImportError:
            # python-drmaa is not installed.
            logger.error(
                "Was not able to import drmaa. " +
                "If this is meant to be run using the OpenGrid submission " +
                "system, then drmaa needs to be installed via pip or " +
                "easy_install."
            )
            raise
        except RuntimeError:
            # The drmaa library was not specified, but python-drmaa is
            # installed.
            logger.error(
                "Was able to import drmaa. " +
                "However, the drmaa library could not be found. Please " +
                "either specify the location of libdrmaa.so using the " +
                "DRMAA_LIBRARY_PATH environment variable or disable/remove " +
                "use_drmaa from the config file."
            )
            raise

        s=drmaa.Session()
        s.initialize()

        ready_processes = []
        for each_arg_pack in block_process_args_gen:
            ready_processes.append((each_arg_pack, s.createJobTemplate()))
            ready_processes[-1][1].jobName = os.path.basename(
                os.path.splitext(cur_module_filepath)[0]
            ) + "-" + os.path.basename(
                os.path.dirname(each_arg_pack[3].split(".h5")[0])
            ) + "-" + os.path.basename(each_arg_pack[3].split(".h5")[0])
            ready_processes[-1][1].remoteCommand = each_arg_pack[0]
            ready_processes[-1][1].args = each_arg_pack[1:-2]
            ready_processes[-1][1].jobEnvironment = os.environ
            ready_processes[-1][1].inputPath = "localhost:" + os.devnull
            ready_processes[-1][1].outputPath = "localhost:" + each_arg_pack[-2]
            ready_processes[-1][1].errorPath = "localhost:" + each_arg_pack[-1]
            ready_processes[-1][1].workingDirectory = os.getcwd()
            ready_processes[-1][1].nativeSpecification = "-pe batch " + str(num_drmaa_cores)


        running_processes = []
        for each_arg_pack, each_process_template in ready_processes:
            each_process_id = s.runJob(each_process_template)
            running_processes.append(
                (each_arg_pack, each_process_id, each_process_template)
            )
            logger.info(
                "Started new process ( \"" + " ".join(each_arg_pack) + "\" )."
            )

        start_queue_time = time.time()
        logger.info("Waiting for queued jobs to complete.")

        #finished_processes = []
        for each_arg_pack, each_process_id, each_process_template in running_processes:
            each_process_status = s.wait(each_process_id)

            if not each_process_status.hasExited:
                raise RuntimeError(
                    "The process (\"" + " ".join(each_arg_pack) +
                    "\") has exited prematurely."
                )

            logger.info(
                "Finished process ( \"" + " ".join(each_arg_pack) + "\" )."
            )
            s.deleteJobTemplate(each_process_template)
            #finished_processes.append((each_arg_pack, each_process_id))

        s.exit()

        end_queue_time = time.time()
        diff_queue_time = end_queue_time - start_queue_time

        logger.info(
            "Run time for queued jobs to complete is \""
            + str(diff_queue_time) + " s\"."
        )
    else:
        # TODO: Refactor into a separate class (have it return futures somehow)
        #finished_processes = []
        running_processes = []
        pool_tasks_empty = False
        while (not pool_tasks_empty) or len(running_processes):
            while (not pool_tasks_empty) and (len(running_processes) < num_processes):
                try:
                    each_arg_pack = next(block_process_args_gen)
                    each_arg_pack, each_stdout_filename, each_stderr_filename = each_arg_pack[:-2], each_arg_pack[-2], each_arg_pack[-1]
                    each_process = subprocess.Popen(
                        each_arg_pack,
                        stdout=open(each_stdout_filename, "w"),
                        stderr=open(each_stderr_filename, "w")
                    )

                    running_processes.append((each_arg_pack, each_process,))

                    logger.info(
                        "Started new process ( \"" + " ".join(each_arg_pack) + "\" )."
                    )
                except StopIteration:
                    pool_tasks_empty = True

            while ((not pool_tasks_empty) and
                       (len(running_processes) >= num_processes)) or \
                    (pool_tasks_empty and len(running_processes)):
                time.sleep(1)

                i = 0
                while i < len(running_processes):
                    if running_processes[i][1].poll() is not None:
                        logger.info(
                            "Finished process ( \"" +
                            " ".join(running_processes[i][0]) + "\" )."
                        )

                        #finished_processes.append(running_processes[i])
                        del running_processes[i]
                    else:
                        time.sleep(1)
                        i += 1

        # finished_processes = None

    start_time = time.time()
    logger.info("Starting merge over all blocks.")

    with h5py.File(output_filename_details.externalPath, "a") as output_file_handle:
        output_group = output_file_handle[output_group_name]

        new_neurons_set = segment.get_empty_neuron(
            shape=tuple(original_images_shape_array[1:]), dtype=float
        )

        for i, i_str, (output_filename_block_i, sequential_block_i) in iters.filled_stringify_enumerate(
                iters.izip(output_filename_block, original_images_pared_slices.flat)):
            windowed_slice_i = tuple(
                slice(_1, _2, 1) for _1, _2 in [(None, None)] + sequential_block_i["windowed_stack_selection"].tolist()[1:]
            )
            window_trimmed_i = tuple(
                slice(_1, _2, 1) for _1, _2 in sequential_block_i["windowed_block_selection"].tolist()
            )
            output_filename_block_i = output_filename_block_i.rstrip("/")

            with h5py.File(output_filename_block_i, "r") as each_block_file_handle:
                if "neurons" in each_block_file_handle:
                    neurons_block_i_smaller = hdf5.serializers.read_numpy_structured_array_from_HDF5(
                        each_block_file_handle, "/neurons"
                    )

                    neurons_block_i_windowed_count = numpy.squeeze(
                        numpy.apply_over_axes(
                            numpy.sum,
                            neurons_block_i_smaller["mask"].astype(float),
                            tuple(iters.irange(1, neurons_block_i_smaller["mask"].ndim))
                        )
                    )

                    if neurons_block_i_windowed_count.shape == tuple():
                        neurons_block_i_windowed_count = numpy.array(
                            [neurons_block_i_windowed_count])

                    neurons_block_i_non_windowed_count = numpy.squeeze(
                        numpy.apply_over_axes(
                            numpy.sum,
                            neurons_block_i_smaller["mask"][window_trimmed_i].astype(float),
                            tuple(iters.irange(1, neurons_block_i_smaller["mask"].ndim))
                        )
                    )

                    if neurons_block_i_non_windowed_count.shape == tuple():
                        neurons_block_i_non_windowed_count = numpy.array(
                            [neurons_block_i_non_windowed_count]
                        )

                    if len(neurons_block_i_non_windowed_count):
                        # Find ones that are inside the margins by more than
                        # half
                        neurons_block_i_acceptance = (
                            (neurons_block_i_non_windowed_count / neurons_block_i_windowed_count) > 0.5
                        )

                        logger.info(
                            "Accepted the following neurons %s from block %s."
                            % (
                                str(neurons_block_i_acceptance.nonzero()[0].tolist()),
                                i_str
                            )
                        )

                        # Take a subset of our previous neurons that are within
                        # the margins by half
                        neurons_block_i_accepted = neurons_block_i_smaller[neurons_block_i_acceptance]

                        neurons_block_i = numpy.zeros(
                            neurons_block_i_accepted.shape, dtype=new_neurons_set.dtype
                        )
                        neurons_block_i["mask"][windowed_slice_i] = neurons_block_i_accepted["mask"]
                        neurons_block_i["contour"][windowed_slice_i] = neurons_block_i_accepted["contour"]
                        neurons_block_i["image"][windowed_slice_i] = neurons_block_i_accepted["image"]

                        # Copy other properties
                        neurons_block_i["area"] = neurons_block_i_accepted["area"]
                        neurons_block_i["max_F"] = neurons_block_i_accepted["max_F"]
                        neurons_block_i["gaussian_mean"] = neurons_block_i_accepted["gaussian_mean"]
                        neurons_block_i["gaussian_cov"] = neurons_block_i_accepted["gaussian_cov"]
                        # TODO: Correct centroid to larger block position.
                        neurons_block_i["centroid"] = neurons_block_i_accepted["centroid"]
                        neurons_block_i["centroid"] += sequential_block_i["windowed_stack_selection"][1:, 0]

                        array_debug_recorder = hdf5.record.generate_HDF5_array_recorder(
                            output_group,
                            group_name="debug",
                            enable=debug,
                            overwrite_group=False,
                            recorder_constructor=hdf5.record.HDF5EnumeratedArrayRecorder
                        )

                        segment.merge_neuron_sets.recorders.array_debug_recorder = array_debug_recorder
                        new_neurons_set = segment.merge_neuron_sets(
                            new_neurons_set,
                            neurons_block_i,
                            **parameters["generate_neurons"]["postprocess_data"]["merge_neuron_sets"]
                        )
                    else:
                        logger.info(
                            "Accepted the following neurons %s from block %s." %
                            (
                                str([]),
                                i_str
                            )
                        )
                else:
                    logger.info(
                        "No neurons accepted as none were found for block"
                        " %s." %
                        i_str
                    )

        hdf5.serializers.create_numpy_structured_array_in_HDF5(
            output_group, "neurons", new_neurons_set, overwrite=True)

        if "parameters" not in output_group["neurons"].attrs:
            output_group["neurons"].attrs["parameters"] = repr(dict(
                list(parameters.items()) +
                [("block_shape", block_shape),
                 ("num_blocks", num_blocks),
                 ("half_window_shape", half_window_shape),
                 ("half_border_shape", half_border_shape),
                 ("use_drmaa", use_drmaa),
                 ("num_drmaa_cores", num_drmaa_cores),
                 ("debug", debug)]
            ))

    logger.info("Finished merge over all blocks.")
    end_time = time.time()

    diff_time = end_time - start_time

    logger.info(
        "Run time for merge over all blocks is \"" + str(diff_time) + " s\"."
    )


@prof.log_call(trace_logger)
@hdf5.record.static_subgrouping_array_recorders(array_debug_recorder=hdf5.record.EmptyArrayRecorder())
@wrappers.static_variables(resume_logger=hdf5.record.EmptyArrayRecorder())
def generate_neurons(original_images, run_stage="all", **parameters):
    if "original_images_max_projection" not in generate_neurons.recorders.array_debug_recorder:
        generate_neurons.recorders.array_debug_recorder["original_images_max_projection"] = xnumpy.add_singleton_op(
            numpy.max,
            original_images,
            axis=0
        )

    if "original_images_mean_projection" not in generate_neurons.recorders.array_debug_recorder:
        generate_neurons.recorders.array_debug_recorder["original_images_mean_projection"] = xnumpy.add_singleton_op(
            numpy.mean,
            original_images,
            axis=0
        )

    # Preprocess images
    new_preprocessed_images = generate_neurons.resume_logger.get(
        "preprocessed_images", None
    )
    if (new_preprocessed_images is None) or \
            (run_stage == "preprocessing") or \
            (run_stage == "all"):
        new_preprocessed_images = original_images.copy()
        segment.preprocess_data.recorders.array_debug_recorder = generate_neurons.recorders.array_debug_recorder
        new_preprocessed_images = segment.preprocess_data(
            new_preprocessed_images,
            out=new_preprocessed_images,
            **parameters["preprocess_data"]
        )
        generate_neurons.resume_logger["preprocessed_images"] = new_preprocessed_images

        if "preprocessed_images_max_projection" not in generate_neurons.recorders.array_debug_recorder:
            generate_neurons.recorders.array_debug_recorder["preprocessed_images_max_projection"] = xnumpy.add_singleton_op(
            numpy.max,
            new_preprocessed_images,
            axis=0
        )

    if run_stage == "preprocessing":
        return

    # Find the dictionary
    new_dictionary = generate_neurons.resume_logger.get("dictionary", None)
    if (new_dictionary is None) or \
            (run_stage == "dictionary") or \
            (run_stage == "all"):
        segment.generate_dictionary.recorders.array_debug_recorder = generate_neurons.recorders.array_debug_recorder
        new_dictionary = segment.generate_dictionary(
            new_preprocessed_images,
            **parameters["generate_dictionary"]
        )
        generate_neurons.resume_logger["dictionary"] = new_dictionary

        if "dictionary_max_projection" not in generate_neurons.recorders.array_debug_recorder:
            generate_neurons.recorders.array_debug_recorder["dictionary_max_projection"] = xnumpy.add_singleton_op(
                numpy.max,
                new_dictionary,
                axis=0
            )

    if run_stage == "dictionary":
        return

    # Find the neurons
    new_neurons = None
    new_neurons = generate_neurons.resume_logger.get("neurons", None)
    if (new_neurons is None) or \
            (run_stage == "postprocessing") or \
            (run_stage == "all"):
        segment.postprocess_data.recorders.array_debug_recorder = generate_neurons.recorders.array_debug_recorder
        new_neurons = segment.postprocess_data(
            new_dictionary,
            **parameters["postprocess_data"]
        )

        if new_neurons.size:
            generate_neurons.resume_logger["neurons"] = new_neurons

    if new_neurons.size == 0:
        logger.warning("No neurons were found in the data.")
    else:
        logger.info(
            "Found \"" + str(len(new_neurons)) +
            "\" neurons were found in the data."
        )


@prof.log_call(trace_logger)
def main(*argv):
    """
        Simple main function (like in C). Takes all arguments (as from
        sys.argv) and returns an exit status.

        Args:
            argv(list):     arguments (includes command line call).

        Returns:
            int:            exit code (0 if success)
    """

    # Only necessary if running main (normally if calling command line).
    # No point in importing otherwise.
    import argparse
    import threading

    memory_profiler_thread = threading.Thread(
        target=prof.memory_profiler, args=(logger,)
    )
    memory_profiler_thread.daemon = True
    memory_profiler_thread.start()

    argv = list(argv)

    # Creates command line parser
    parser = argparse.ArgumentParser(
        description="Parses input from the command line for a batch job."
    )

    # Takes a config file and then a series of one or more HDF5 files.
    parser.add_argument(
        "config_filename",
        metavar="CONFIG_FILE",
        type=str,
        help="JSON file that provides configuration options " +
             "for how to use dictionary learning on the input files."
    )
    parser.add_argument(
        "input_file",
        metavar="INPUT_FILE",
        type=str,
        nargs=1,
        help="HDF5 file with an array of images. " +
             "A dataset or video will be expected at the internal path. " +
             "Time must be the first dimension."
    )
    parser.add_argument(
        "output_file",
        metavar="OUTPUT_FILE",
        type=str,
        nargs=1,
        help="HDF5 file(s) to write output. " +
             "If a specific group is desired, " +
             "that should be included in the filename."
    )

    # Results of parsing arguments
    # (ignore the first one as it is the command line call).
    parsed_args = parser.parse_args(argv[1:])

    # Remove args from singleton lists
    parsed_args.input_file = parsed_args.input_file[0]
    parsed_args.output_file = parsed_args.output_file[0]

    # Runs the dictionary learning algorithm on each file with
    # the given parameters and saves the results in the given files.
    generate_neurons_io_handler(
        parsed_args.input_file,
        parsed_args.output_file,
        parsed_args.config_filename
    )

    return(0)



"""
The ``masks`` module provides filters for working with binary images.

===============================================================================
Overview
===============================================================================
Functions provided include binary dilation and erosion. These allow for N-D
array processing.

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Mar 31, 2015 20:47:37 EDT$"


import numpy

import scipy
import scipy.ndimage
import scipy.ndimage.filters

# Need in order to have logging information no matter what.
from nanshe.util import prof


# Get the logger
trace_logger = prof.getTraceLogger(__name__)


@prof.log_call(trace_logger)
def binary_dilation(input_array, footprint, out=None):
    """
        Performs simple binary dilation on a bool array of arbitrary dimension.

        Args:
            input_array(numpy.ndarray):         the bool array to perform
                                                dilation on.

            footprint(numpy.ndarray):           the footprint to use for the
                                                kernel.

            out(numpy.ndarray):                 a place to store the result if
                                                provided. (optional)

        Returns:
            out(numpy.ndarray):                 Same as out if out was
                                                provided.

        >>> a = numpy.array(
        ...     [[ True,  True, False, False, False, False, False],
        ...      [False, False, False, False, False, False, False],
        ...      [False, False, False, False, False, False, False],
        ...      [False, False, False, False, False, False, False],
        ...      [False, False, False, False,  True, False, False],
        ...      [False, False, False, False, False, False, False],
        ...      [False, False, False, False, False, False, False]], dtype=bool
        ... )
        >>> b = numpy.zeros_like(a)

        >>> binary_dilation(a, numpy.ones(a.ndim*(3,), dtype=bool))
        array([[ True,  True,  True, False, False, False, False],
               [ True,  True,  True, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False,  True,  True,  True, False],
               [False, False, False,  True,  True,  True, False],
               [False, False, False,  True,  True,  True, False],
               [False, False, False, False, False, False, False]], dtype=bool)

        >>> binary_dilation(a, numpy.ones(a.ndim*(3,), dtype=bool), out=b)
        array([[ True,  True,  True, False, False, False, False],
               [ True,  True,  True, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False,  True,  True,  True, False],
               [False, False, False,  True,  True,  True, False],
               [False, False, False,  True,  True,  True, False],
               [False, False, False, False, False, False, False]], dtype=bool)
        >>> b
        array([[ True,  True,  True, False, False, False, False],
               [ True,  True,  True, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False,  True,  True,  True, False],
               [False, False, False,  True,  True,  True, False],
               [False, False, False,  True,  True,  True, False],
               [False, False, False, False, False, False, False]], dtype=bool)

        >>> binary_dilation(a, numpy.ones(a.ndim*(3,), dtype=bool), out=a)
        array([[ True,  True,  True, False, False, False, False],
               [ True,  True,  True, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False,  True,  True,  True, False],
               [False, False, False,  True,  True,  True, False],
               [False, False, False,  True,  True,  True, False],
               [False, False, False, False, False, False, False]], dtype=bool)
        >>> a
        array([[ True,  True,  True, False, False, False, False],
               [ True,  True,  True, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False,  True,  True,  True, False],
               [False, False, False,  True,  True,  True, False],
               [False, False, False,  True,  True,  True, False],
               [False, False, False, False, False, False, False]], dtype=bool)
    """

    assert issubclass(input_array.dtype.type, (bool, numpy.bool_))

    if out is None:
        out = input_array.copy()
    elif id(input_array) != id(out):
        assert issubclass(out.dtype.type, (bool, numpy.bool_))

    scipy.ndimage.filters.maximum_filter(
        input_array, footprint=footprint, output=out
    )

    return(out)


@prof.log_call(trace_logger)
def binary_erosion(input_array, footprint, out=None):
    """
        Performs simple binary erosion on a bool array of arbitrary dimension.

        Args:
            input_array(numpy.ndarray):         the bool array to perform
                                                erosion on.

            footprint(numpy.ndarray):           the footprint to use for the
                                                kernel.

            out(numpy.ndarray):                 a place to store the result if
                                                provided. (optional)

        Returns:
            out(numpy.ndarray):                 Same as out if out was
                                                provided.

        >>> a = numpy.array(
        ...     [[ True,  True,  True, False, False, False,  True],
        ...      [ True,  True,  True, False, False,  True,  True],
        ...      [False, False, False, False, False, False, False],
        ...      [False, False, False,  True,  True,  True, False],
        ...      [False,  True, False,  True,  True,  True, False],
        ...      [False,  True, False,  True,  True,  True, False],
        ...      [False, False, False, False, False, False,  True]], dtype=bool
        ... )
        >>> b = numpy.zeros_like(a)

        >>> binary_erosion(a, numpy.ones(a.ndim*(3,), dtype=bool))
        array([[ True,  True, False, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False,  True, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False, False, False, False]], dtype=bool)

        >>> binary_erosion(a, numpy.ones(a.ndim*(3,), dtype=bool), out=b)
        array([[ True,  True, False, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False,  True, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False, False, False, False]], dtype=bool)
        >>> b
        array([[ True,  True, False, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False,  True, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False, False, False, False]], dtype=bool)

        >>> binary_erosion(a, numpy.ones(a.ndim*(3,), dtype=bool), out=a)
        array([[ True,  True, False, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False,  True, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False, False, False, False]], dtype=bool)
        >>> a
        array([[ True,  True, False, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False,  True, False, False],
               [False, False, False, False, False, False, False],
               [False, False, False, False, False, False, False]], dtype=bool)
    """

    assert issubclass(input_array.dtype.type, (bool, numpy.bool_))

    if out is None:
        out = input_array.copy()
    elif id(input_array) != id(out):
        assert issubclass(out.dtype.type, (bool, numpy.bool_))

    scipy.ndimage.filters.minimum_filter(
        input_array, footprint=footprint, output=out
    )

    return(out)



"""
``nanshe`` package, an image processing toolkit.

===============================================================================
Overview
===============================================================================

The ``nanshe`` package is an image processing package that contains a variety
of different techniques, which are used primarily to assemble the ADINA
algorithm proposed by Diego, et al.
( doi:`10.1109/ISBI.2013.6556660`_ ) to extract active neurons from
an image sequence. This algorithm uses online dictionary learning (a form of
matrix factorization) at its heart as implemented by Marial, et al.
( doi:`10.1145/1553374.1553463`_ ) to find a set of atoms (or basis
images) that are representative of an image sequence and can be used to
approximately reconstruct the sequence. However, it is designed in a modular
way so that a different matrix factorization could be swapped in and
appropriately parameterized. Other portions of the algorithm include a
preprocessing phase that has a variety of different techniques that can be
applied optionally. For example, removing registration artifacts from
a line-by-line registration algorithm, background subtraction, and a wavelet
transform to filter objects in a particular size.

===============================================================================
 Installation
===============================================================================

-------------------------------------------------------------------------------
 Dependencies
-------------------------------------------------------------------------------
Implementation of the algorithm has been done here in pure Python. However, a
few dependencies are required to get started. These include NumPy_, SciPy_,
h5py_, scikit-image_, SPAMS_, VIGRA_, and rank_filter_. The first 4 can be
found in standard distributions like Anaconda_. Installing VIGRA and
rank_filter can be done by using CMake_. SPAMS requires an existing BLAS/LAPACK
implementation. On Mac and Linux, this can be anything. Typically ATLAS_ is
used, but OpenBLAS_ or `Intel MKL`_ (if available) can be used, as well. This
will require modifying the setup.py script. On Windows, the setup.py links to
R_, which should be changed if another BLAS is available.

-------------------------------------------------------------------------------
 Building
-------------------------------------------------------------------------------
Python
===============================================================================
As this package is pure Python, building follows through the standard method.
Currently, we require setuptools_ for installation; so, make sure it is
installed. Then simply issue the following command to build and install.

.. code-block:: sh

    python setup.py install

Alternatively, one can build and then install in two steps if that is
preferable.

.. code-block:: sh

    python setup.py build
    python setup.py install

Conda
===============================================================================
Current packages can be found on our anaconda_ channel
( https://anaconda.org/nanshe/nanshe ). New ones are released every time a
passing tagged release is pushed to the ``master`` branch on GitHub. It is also
possible to build packages for conda_ for non-release commits as we do in our
continuous integration strategy.

To do this one requires the dependencies be installed or be available from a
anaconda channel. Additionally, one must be using the conda's ``root``
environment and have conda-build installed. Once this is done one need
only the run the following command with ``setup.py``.

.. code-block:: sh

    python setup.py bdist_conda

Assuming this completes successfully, conda will provide the path to the built
package.

-------------------------------------------------------------------------------
 Testing
-------------------------------------------------------------------------------
Running the test suite is fairly straightforward. Testing is done using nose_;
so, make sure you have a running copy if you wish to run the tests. Some of the
tests require drmaa_ installed and properly configured. If that is not the
case, those tests will be skipped automatically. To run the test suite, one
must be in the source directory. Then simply run the following command. This
will run all the tests and doctests. Depending on your machine, this will take
a few minutes to complete.

.. code-block:: sh

    nosetests

The full test suite includes 3D tests, which are very slow to run and so are
not run by default. As the code has been written to be dimensionally agnostic,
these tests don't cover anything that the 2D tests don't already cover. To run
the 3D tests, simply use ``setup.all.cfg``.

.. code-block:: sh

    nosetests -c setup.all.cfg

It is also possible to run this as part of the setup.py process. In which case,
this can be done as shown below. If 3D tests are required for this portion, one
need only replace ``setup.cfg`` with ``setup.all.cfg``.

.. code-block:: sh

    python setup.py nosetests

Also, the typical ``test`` subcommand can be used to run ``nosetests``, but no
other arguments are allowed.

.. code-block:: sh

    python setup.py test

-------------------------------------------------------------------------------
 Documentation
-------------------------------------------------------------------------------
Current documentation can be found on the GitHub page
( http://nanshe-org.github.io/nanshe/ ). A new copy is rebuilt any time there is
a passing commit is added to the ``master`` branch. Each documentation commit
is added to ``gh-pages`` branch with a reference to the commit in ``master``
that triggered the build as well as the tag (version) if provided.

It is also possible to build the documentation from source. This project uses
Sphinx_ for generating documentation. Please make sure you have it installed.
In particular, a version from 1.3 or later is required. Additionally, the
`Cloud Sphinx Theme`_ is required for generating the documentation and is used
in the HTML layout.

The ``rst`` files (outside of ``index.rst`` are not distributed with the source
code. This is because it is trivial to generate them and it is to easy for the
code to become out of sync with documentation if they are distributed. However,
building ``rst`` files has been made a dependency of all other documentation
build steps so one does not have to think about this. The preferred method for
building documentation is to use the ``setup.py`` hook as shown below. This
will build the RST files and place them in ``docs/``. It will also build the
HTML files by default and put them in the directory ``build/sphinx/html/``.
Simply open the ``index.html`` file to take a look.

.. code-block:: sh

    python setup.py build_sphinx

More build options can be determined by running the help command.

.. code-block:: sh

    python setup.py build_sphinx --help

-------------------------------------------------------------------------------
 Cleaning
-------------------------------------------------------------------------------
After any building operation a number of undesirable intermediate files are
created and left behind that one may wish to remove. To do this one merely
needs to run the clean command.

.. code-block:: sh

    python setup.py clean

This has been modified to also remove RST files generated when building
documentation. However, it will leave any final build products like HTML files.
If one wishes to remove everything built (including final build products), the
clean all command will do this.

.. code-block:: sh

    python setup.py clean --all

.. _`10.1109/ISBI.2013.6556660`: http://dx.doi.org/10.1109/ISBI.2013.6556660
.. _`10.1145/1553374.1553463`: http://dx.doi.org/10.1145/1553374.1553463
.. _NumPy: http://www.numpy.org/
.. _SciPy: http://www.scipy.org/
.. _h5py: http://www.h5py.org/
.. _scikit-image: http://scikit-image.org/
.. _SPAMS: http://spams-devel.gforge.inria.fr/
.. _VIGRA: http://ukoethe.github.io/vigra/
.. _rank_filter: http://github.com/nanshe-org/rank_filter/
.. _Anaconda: http://store.continuum.io/cshop/anaconda/
.. _CMake: http://www.cmake.org/
.. _ATLAS: http://math-atlas.sourceforge.net/
.. _OpenBLAS: http://www.openblas.net/
.. _`Intel MKL`: http://software.intel.com/en-us/intel-mkl
.. _R: http://www.r-project.org/
.. _setuptools: http://pythonhosted.org/setuptools/
.. _anaconda: https://anaconda.org/
.. _conda: http://conda.pydata.org/
.. _nose: http://nose.readthedocs.org/en/latest/
.. _drmaa: http://github.com/pygridtools/drmaa-python
.. _Sphinx: http://sphinx-doc.org/
.. _`Cloud Sphinx Theme`: https://pythonhosted.org/cloud_sptheme/
"""
__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Dec 22, 2014 08:46:12 EST$"


from ._version import get_versions
__version__ = get_versions()['version']
del get_versions


__all__ = [
    "box", "converter", "io", "imp", "learner", "registerer", "syn", "util",
    # "viewer"
]

from nanshe import box
from nanshe import converter
from nanshe import io
from nanshe import imp
from nanshe import learner
from nanshe import registerer
from nanshe import syn
from nanshe import util
# from nanshe import viewer    # Must be commented as there is some segfault
                               # coming from Volumina.



"""
The package ``syn`` provides components necessary to generate synthetic data.

===============================================================================
Overview
===============================================================================
The package ``syn`` (short for **syn**\ thetic) contains components necessary
for generating synthetic data. This is primarily used for testing the
algorithm. However, it could be used for benchmarking algorithms or in other
cases.
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Oct 15, 2014 18:16:33 EDT$"

__all__ = ["data"]


from nanshe.syn import data



"""
The module ``prof`` provides support tracing and profiling.

===============================================================================
Overview
===============================================================================
The module ``prof`` provides a few primitives for tracing function calls and
memory usage. In particular, it provides a trace logger that gives us feedback
about arguments passed, the run time, the exception raised, etc. Decorators are
available for wrapping functions and classes (all methods). The special case of
``Qt`` inheriting classes is handled by a separate decorator. In addition to
the trace logger, a memory profiler is also provided, which can be run in a
separate thread to get information about memory usage. The memory profiler
requires |psutil|_.

.. |psutil| replace:: ``psutil``
.. _psutil: http://github.com/giampaolo/psutil

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Apr 14, 2014 17:42:07 EDT$"


import logging
import traceback
import os
import sys
import time

import psutil

from nanshe.util import wrappers



# Nothing fancy. Just the basic logging
# unless otherwise specified, in which case this does nothing.
logging.basicConfig(level=logging.DEBUG, stream=sys.stderr)


def getSpecialLogger(logger_prefix, name, *args, **kwargs):
    """
        A fancy version of ``logging.getLogger``, which takes a prefix and a
        name, which it joins together as the returned logger's name.

        Args:
            logger_prefix(str):         Prefix name to use to describe all
                                        loggers of this type.

            name(str):                  The name of the function or module
                                        being logged.

            *args:                      Other arguments to pass through to
                                        ``getLogger``. Currently, it takes no
                                        others.

            *kwargs:                    Other keyword arguments to pass through
                                        to ``getLogger``. Currently, it takes
                                        no others.

        Returns:
            logging.Logger:             A logger with the given prefix and
                                        subsequent name.
    """

    return(
        logging.getLogger(".".join([logger_prefix, name]), *args, **kwargs)
    )


def getTraceLogger(name, *args, **kwargs):
    """
        A fancy version of ``logging.getLogger``, which adds the prefix TRACE
        to the name given.

        Args:
            name(str):                  The name of the function or module
                                        being logged.

            *args:                      Other arguments to pass through to
                                        ``getLogger``. Currently, it takes no
                                        others.

            *kwargs:                    Other keyword arguments to pass through
                                        to ``getLogger``. Currently, it takes
                                        no others.

        Returns:
            logging.Logger:             A logger with the given name.
    """

    return(getSpecialLogger("TRACE", name, *args, **kwargs))


def getTraceMetaLogger(name, *args, **kwargs):
    """
        A fancy version of ``logging.getLogger``, which adds the prefix
        TRACE.META to the name given.

        Args:
            name(str):                  The name of the function or module
                                        being logged.

            *args:                      Other arguments to pass through to
                                        ``getLogger``. Currently, it takes no
                                        others.

            *kwargs:                    Other keyword arguments to pass through
                                        to ``getLogger``. Currently, it takes
                                        no others.

        Returns:
            logging.Logger:             A logger with the given name.
    """
    return(getSpecialLogger("TRACE.META", name, *args, **kwargs))


def log_call(logger,
             to_log_call=True,
             to_print_args=False,
             to_print_time=True,
             to_print_exception=False):
    """
        Takes a given logger and uses it to log entering and leaving the
        decorated callable. Intended to be used as a decorator that takes a few
        arguments.

        Args:
            logger(Logger):                Used for logging entry, exit and
                                           possibly arguments.

        Keyword Args:
            to_log_call(bool):             Whether to log call or not. This
                                           overrides all other arguments. It
                                           will be stored as a global variable
                                           on the function, which can be
                                           changed at runtime.

            to_print_args(bool):           Whether to output the arguments and
                                           keyword arguments passed to the
                                           function. This should not
                                           automatically be true as some
                                           arguments may not be printable or
                                           may be expensive to print. Thus, it
                                           should be up to the developer to use
                                           their own discretion. Further, we
                                           don't want to break their existing
                                           code. It will be stored as a global
                                           variable on the function, which can
                                           be changed at runtime.

            to_print_time(bool):           Prints the time it took to run the
                                           wrapped callable.

            to_print_exception(bool):      Whether to print the traceback when
                                           an exception is raise. It will be
                                           stored as a global variable on the
                                           function, which can be changed at
                                           runtime.

        Returns:
            log_call_decorator:            For performing the actual wrapping.
    """

    def log_call_decorator(callable):
        """
            The actual decorator, which is what is returned to decorate the
            callable in question.

            Args:
                callable:                              Anything that can be
                                                       called function, method,
                                                       functor, lambda, etc.

            Returns:
                log_call_callable_wrapped(callable):   which is wrapped around
                                                       the function in
                                                       question.
        """

        @wrappers.wraps(callable)
        @wrappers.static_variables(to_log_call=to_log_call,
                                   to_print_args=to_print_args,
                                   to_print_time=to_print_time,
                                   to_print_exception=to_print_exception)
        def log_call_callable_wrapped(*args, **kwargs):
            """
                This is what will replace the original callable. It will behave
                the same except it will now log its entrance and exit. If set,
                it will log its parameters too.

                Args:
                    callable:                               Anything that can
                                                            be called function,
                                                            method, functor,
                                                            lambda, etc.

                Returns:
                    log_call_callable_wrapped (callable):   which is wrapped
                                                            around the function
                                                            in question.
            """

            result = None
            # This allows keyword arguments to be turned on or off at runtime.
            if log_call_callable_wrapped.to_log_call:
                # Log that we have entered the callable in question.
                logger.debug(
                    "Entering callable: \"" + callable.__name__ + "\"."
                )

                # Output arguments and keyword arguments if acceptable.
                # This allows keyword arguments to be turned on or off at
                # runtime.
                #
                # Note: We have used log_call_callable_wrapped.to_print_args.
                # However, we cannot define this until after as wrapping will
                # lose this variable.
                if (log_call_callable_wrapped.to_print_args):
                    logger.debug(
                        "Arguments: \"" + str(args) + "\"" + os.linesep +
                        "Keyword Arguments: \"" + str(kwargs) + "\"."
                    )

                # We don't return immediately. Why? We want to know if this
                # succeeded or failed. So, we want the log message below to
                # print after the function runs.
                diff_time = 0.0
                start_time = time.time()
                try:
                    result = callable(*args, **kwargs)
                except:
                    if log_call_callable_wrapped.to_print_exception:
                        logger.error(traceback.format_exc())
                    raise
                end_time = time.time()
                diff_time += (end_time - start_time)

                # Log that we have exited the callable in question.
                logger.debug(
                    "Exiting callable: \"" + callable.__name__ + "\"."
                )

                if log_call_callable_wrapped.to_print_time:
                    logger.debug(
                        "Run time for callable: \"" + callable.__name__ +
                        "\" is \"" + str(diff_time) + " s\"."
                    )
            else:
                result = callable(*args, **kwargs)

            # Return the result even None.
            return(result)

        # The callable wrapped.
        return(log_call_callable_wrapped)

    # The arguments passed to the decorator for easy access.
    return(log_call_decorator)


def log_class(logger,
              to_log_call=True,
              to_print_args=False,
              to_print_time=True,
              to_print_exception=False):
    """
        Takes a given logger and uses it to log entering and leaving all
        methods of the decorated class. Intended to be used as a decorator that
        takes a few arguments.

        Args:
            logger(Logger):                Used for logging entry, exit and
                                           possibly arguments.

        Keyword Args:
            to_log_call(bool):             Whether to log call or not. This
                                           overrides all other arguments. It
                                           will be stored as a global variable
                                           on the methods, which can be changed
                                           at runtime.

            to_print_args(bool):           Whether to output the arguments and
                                           keyword arguments passed to the
                                           function. This should not
                                           automatically be true as some
                                           arguments may not be printable or
                                           may be expensive to print. Thus, it
                                           should be up to the developer to use
                                           their own discretion. Further, we
                                           don't want to break their existing
                                           code. It will be stored as a global
                                           variable on the methods, which can
                                           be changed at runtime.

            to_print_time(bool):           Prints the time it took to run the
                                           wrapped callable.

            to_print_exception(bool):      Whether to print the traceback when
                                           an exception is raise. It will be
                                           stored as a global variable on the
                                           methods, which can be changed at
                                           runtime.

        Returns:
            log_call_decorator(for wrapping)
    """

    return(wrappers.class_decorate_all_methods(log_call(
        logger,
        to_log_call=to_log_call,
        to_print_args=to_print_args,
        to_print_time=to_print_time,
        to_print_exception=to_print_exception
    )))


def qt_log_class(logger,
                 to_log_call=True,
                 to_print_args=False,
                 to_print_time=True,
                 to_print_exception=False):
    """
        Takes a given logger and uses it to log entering and leaving all
        methods of the decorated class. Intended to be used as a decorator that
        takes a few arguments.

        Args:
            logger(Logger):                Used for logging entry, exit and
                                           possibly arguments.

        Keyword Args:
            to_log_call(bool):             Whether to log call or not. This
                                           overrides all other arguments. It
                                           will be stored as a global variable
                                           on the methods, which can be changed
                                           at runtime.

            to_print_args(bool):           Whether to output the arguments and
                                           keyword arguments passed to the
                                           function. This should not
                                           automatically be true as some
                                           arguments may not be printable or
                                           may be expensive to print. Thus, it
                                           should be up to the developer to use
                                           their own discretion. Further, we
                                           don't want to break their existing
                                           code. It will be stored as a global
                                           variable on the methods, which can
                                           be changed at runtime.

            to_print_time(bool):           Prints the time it took to run the
                                           wrapped callable.

            to_print_exception(bool):      Whether to print the traceback when
                                           an exception is raise. It will be
                                           stored as a global variable on the
                                           methods, which can be changed at
                                           runtime.

        Returns:
            log_call_decorator(for wrapping)
    """

    return(wrappers.qt_class_decorate_all_methods(log_call(
        logger,
        to_log_call=to_log_call,
        to_print_args=to_print_args,
        to_print_time=to_print_time,
        to_print_exception=to_print_exception
    )))


@wrappers.static_variables(to_run=True)
def memory_profiler(logger, interval=1, level=logging.INFO):
    """
        Runs forever get information about memory usage and dumping it to the
        logger provided at the given interval.

        Args:
            logger(Logger):               Used for logging memory profiling
                                          information.

        Keyword Args:
            interval(int or float):       Number of seconds to wait before
                                          issuing more profile information.
    """

    current_process = psutil.Process(os.getpid())

    while memory_profiler.to_run:
        logger.log(
            level, "Memory info = " + repr(current_process.memory_info_ex())
        )
        time.sleep(interval)



"""
The module ``iters`` provides support working with and creating generators.

===============================================================================
Overview
===============================================================================
The module ``iters`` provides a variety of functions targeted at working with
or creating generators. It also contains some functions targeted at working
with slices and tuples of slices.

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Apr 17, 2014 13:43:56 EDT$"


import itertools
import math

import numpy

# Need in order to have logging information no matter what.
from nanshe.util import prof


try:
    from itertools import izip_longest
except ImportError:
    from itertools import zip_longest as izip_longest

try:
    from itertools import imap
except ImportError:
    imap = map

try:
    from itertools import izip
except ImportError:
    izip = zip

try:
    irange = xrange
except NameError:
    irange = range


# Get the logger
trace_logger = prof.getTraceLogger(__name__)


@prof.log_call(trace_logger)
def index_generator(*sizes):
    """
        Takes an argument list of sizes and iterates through them from 0 up to
        (but not including) each size.

        Args:
            *sizes(int):            an argument list of ints for the max sizes
                                    in each index.

        Returns:
            chain_gen(generator):   a generator over every possible coordinated


        Examples:
            >>> index_generator(0) #doctest: +ELLIPSIS
            <itertools.product object at 0x...>

            >>> list(index_generator(0))
            []

            >>> list(index_generator(0, 2))
            []

            >>> list(index_generator(2, 0))
            []

            >>> list(index_generator(2, 1))
            [(0, 0), (1, 0)]

            >>> list(index_generator(1, 2))
            [(0, 0), (0, 1)]

            >>> list(index_generator(3, 2))
            [(0, 0), (0, 1), (1, 0), (1, 1), (2, 0), (2, 1)]
    """

    # Creates a list of irange generator objects over each respective
    # dimension of sizes
    gens = [irange(_) for _ in sizes]

    # Combines the generators to a single generator of indices that go
    # throughout sizes
    chain_gen = itertools.product(*gens)

    return(chain_gen)


@prof.log_call(trace_logger)
def index_enumerator(*sizes):
    """
        Takes an argument list of sizes and iterates through them from 0 up to
        (but not including) each size (i.e. like index_generator). However,
        also included is an index corresponding to how many elements have been
        seen.

        Args:
            *sizes(int):            an argument list of ints for the max sizes
                                    in each index.

        Returns:
            chain_gen(generator):   a generator over every possible coordinated


        Examples:
            >>> index_enumerator(0) #doctest: +ELLIPSIS
            <enumerate object at 0x...>

            >>> list(index_enumerator(0))
            []

            >>> list(index_enumerator(0, 2))
            []

            >>> list(index_enumerator(2, 0))
            []

            >>> list(index_enumerator(2, 1))
            [(0, (0, 0)), (1, (1, 0))]

            >>> list(index_enumerator(1, 2))
            [(0, (0, 0)), (1, (0, 1))]

            >>> list(index_enumerator(3, 2))
            [(0, (0, 0)), (1, (0, 1)), (2, (1, 0)), (3, (1, 1)), (4, (2, 0)), (5, (2, 1))]
    """

    # Creates a list of irange generator objects over each respective
    # dimension of sizes
    gens = [irange(_) for _ in sizes]

    # Combines the generators to a single generator of indices that go
    # throughout sizes
    chain_gen = enumerate(itertools.product(*gens))

    return(chain_gen)


@prof.log_call(trace_logger)
def list_indices_to_index_array(list_indices):
    """
        Converts a list of tuple indices to numpy index array.

        Args:
            list_indices(list):    a list of indices corresponding to some
                                   array object

        Returns:
            chain_gen(tuple):      a tuple containing a numpy array for each
                                   index

        Examples:
            >>> list_indices_to_index_array([])
            ()

            >>> list_indices_to_index_array([(1,2)])
            (array([1]), array([2]))

            >>> list_indices_to_index_array([(1, 2), (5, 7), (33, 2)])
            (array([ 1,  5, 33]), array([2, 7, 2]))
    """

    # Combines the indices so that one dimension is represented by each list.
    # Then converts this to a tuple numpy.ndarrays.
    return(tuple(numpy.array(list(izip(*list_indices)))))


@prof.log_call(trace_logger)
def list_indices_to_numpy_bool_array(list_indices, shape):
    """
        Much like list_indices_to_index_array except that it constructs
        numpy.ndarray with dtype of bool. All indices in list_indices are set
        to True in the numpy.ndarray. The rest are False by default.

        Args:
            list_indices(list):      a list of indices corresponding to some
                                     numpy.ndarray object

            shape(tuple):            a tuple used to set the shape of the
                                     numpy.ndarray to return

        Returns:
            result(numpy.ndarray):   a numpy.ndarray with dtype bool (True for
                                     indices in list_indices and False
                                     otherwise).

        Examples:
            >>> list_indices_to_numpy_bool_array([], ())
            array(False, dtype=bool)

            >>> list_indices_to_numpy_bool_array([], (0))
            array([], dtype=bool)

            >>> list_indices_to_numpy_bool_array([], (0,0))
            array([], shape=(0, 0), dtype=bool)

            >>> list_indices_to_numpy_bool_array([], shape=(1))
            array([False], dtype=bool)

            >>> list_indices_to_numpy_bool_array([(0,0)], (1,1))
            array([[ True]], dtype=bool)

            >>> list_indices_to_numpy_bool_array(
            ...     [(2,3), (0,0), (0,2), (1,1)], (3,4)
            ... )
            array([[ True, False,  True, False],
                   [False,  True, False, False],
                   [False, False, False,  True]], dtype=bool)
    """

    # Constructs the numpy.ndarray with False everywhere
    result = numpy.zeros(shape, dtype=bool)

    # Gets the index array
    # Done first to make sure that if list_indices is this [], or this (), or this [()]
    # will be converted to this ().
    index_array = list_indices_to_index_array(list_indices)

    # Sets the given indices to True
    if index_array != ():
        result[index_array] = True

    return(result)


@prof.log_call(trace_logger)
def repeat_generator(a_iter, n=1):
    """
        Repeats each value on an iterator given n-times before proceeding to
        the next value.

        Args:
            a_iter(iter):          an iterator that will have its values
                                   repeated contiguously

            n(int):                number of times to repeat each value

        Returns:
            (generator object):    a generator that contiguously repeats value
                                   from the given iterator.

        Examples:
            >>> repeat_generator(irange(5)) #doctest: +ELLIPSIS
            <generator object repeat_generator at 0x...>

            >>> list(repeat_generator(irange(0)))
            []

            >>> list(repeat_generator(irange(5), 0))
            []

            >>> list(repeat_generator(irange(5), 1))
            [0, 1, 2, 3, 4]

            >>> list(repeat_generator(irange(5), 2))
            [0, 0, 1, 1, 2, 2, 3, 3, 4, 4]

            >>> list(repeat_generator(irange(5), 3))
            [0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4]
    """

    assert (n >= 0), "n must be positive, but got n = " + repr(n)
    assert ((n % 1) == 0), "n must be an integer, but got n = " + repr(n)

    for each_value in a_iter:
        for i in irange(n):
            yield(each_value)


@prof.log_call(trace_logger)
def cycle_generator(a_iter, n=1):
    """
        Cycles through an iterator n-times.

        Args:
            a_iter(iter):          an iterator that will be repeated
            n(int):                number of times to repeat the iterator

        Returns:
            (generator object):    a generator that repeats the given iterator
                                   a certain number of times.

        Examples:
            >>> cycle_generator(irange(5)) #doctest: +ELLIPSIS
            <generator object cycle_generator at 0x...>

            >>> list(cycle_generator(irange(0)))
            []

            >>> list(cycle_generator(irange(5), 0))
            []

            >>> list(cycle_generator(irange(5), 1))
            [0, 1, 2, 3, 4]

            >>> list(cycle_generator(irange(5), 2))
            [0, 1, 2, 3, 4, 0, 1, 2, 3, 4]

            >>> list(cycle_generator(irange(5), 3))
            [0, 1, 2, 3, 4, 0, 1, 2, 3, 4, 0, 1, 2, 3, 4]
    """

    assert (n >= 0), "n must be positive, but got n = " + repr(n)
    assert ((n % 1) == 0), "n must be an integer, but got n = " + repr(n)

    a_iter = itertools.chain(*itertools.tee(a_iter, n))

    for each_value in a_iter:
        yield(each_value)


@prof.log_call(trace_logger)
def iter_with_skip_indices(a_iter, to_skip=None):
    """
        Behaves as a normal iterator except allows for skipping arbitrary
        values, as well. These values to be skipped should be specified by
        their indices using some iterable.

        Args:
            a_iter(iter):          an iterator that will skip some values

            to_skip(iter):         some form of iterable or list of indices to
                                   skip (can be a single value as well).

        Returns:
            (generator object):    a generator that skips some values with
                                   indices in to_skip.

        Examples:
            >>> iter_with_skip_indices(irange(10)) #doctest: +ELLIPSIS
            <generator object iter_with_skip_indices at 0x...>

            >>> list(iter_with_skip_indices(irange(10)))
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

            >>> list(iter_with_skip_indices(irange(10), to_skip = 2))
            [0, 1, 3, 4, 5, 6, 7, 8, 9]

            >>> list(iter_with_skip_indices(irange(1, 10), to_skip = 2))
            [1, 2, 4, 5, 6, 7, 8, 9]

            >>> list(iter_with_skip_indices(irange(10), to_skip = [2, 7]))
            [0, 1, 3, 4, 5, 6, 8, 9]

            >>> list(iter_with_skip_indices(irange(10), to_skip = [0]))
            [1, 2, 3, 4, 5, 6, 7, 8, 9]

            >>> list(iter_with_skip_indices(irange(1, 10), to_skip = [0]))
            [2, 3, 4, 5, 6, 7, 8, 9]

            >>> list(iter_with_skip_indices(irange(10), to_skip = [9]))
            [0, 1, 2, 3, 4, 5, 6, 7, 8]
    """

    full = iter(a_iter)
    full_enum = enumerate(full)

    if to_skip is None:
        to_skip = iter([])
    else:
        try:
            to_skip = iter(sorted(set(to_skip)))
        except TypeError:
            to_skip = iter([to_skip])

    next_to_skip = next(to_skip, None)

    for i, each in full_enum:
        if i != next_to_skip:
            yield(each)
        else:
            next_to_skip = next(to_skip, None)


@prof.log_call(trace_logger)
def iter_with_skip_values(a_iter, to_skip=None):
    """
        Behaves as a normal iterator except allows for skipping arbitrary
        values, as well. These values to be skipped should be specified by
        their indices using some iterable.

        Args:
            a_iter(iter):          an iterator that will skip some values

            to_skip(iter):         some form of iterable or list of indices to
                                   skip (can be a single value as well).

        Returns:
            (generator object):    a generator that skips some values with
                                   indices in to_skip.

        Examples:
            >>> iter_with_skip_values(10) #doctest: +ELLIPSIS
            <generator object iter_with_skip_values at 0x...>

            >>> list(iter_with_skip_values(irange(10)))
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

            >>> list(iter_with_skip_values(irange(10)))
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

            >>> list(iter_with_skip_values(irange(1, 10)))
            [1, 2, 3, 4, 5, 6, 7, 8, 9]

            >>> list(iter_with_skip_values(irange(0, 10, 2)))
            [0, 2, 4, 6, 8]

            >>> list(iter_with_skip_values(irange(1, 10, 2)))
            [1, 3, 5, 7, 9]

            >>> list(iter_with_skip_values(irange(10), to_skip = 2))
            [0, 1, 3, 4, 5, 6, 7, 8, 9]

            >>> list(iter_with_skip_values(irange(1, 10), to_skip = 2))
            [1, 3, 4, 5, 6, 7, 8, 9]

            >>> list(iter_with_skip_values(irange(0, 10, 2), to_skip = [2,6]))
            [0, 4, 8]
    """

    full = iter(a_iter)

    if to_skip is None:
        to_skip = []
    else:
        try:
            to_skip = sorted(set(to_skip))
        except TypeError:
            to_skip = [to_skip]

    for each in full:
        if each not in to_skip:
            yield(each)


@prof.log_call(trace_logger)
def xrange_with_skip(start, stop=None, step=None, to_skip=None):
    """
        Behaves as irange does except allows for skipping arbitrary values, as
        well. These values to be skipped should be specified using some
        iterable.

        Args:
            start(int):            start for irange or if stop is not specified
                                   this will be stop.

            stop(int):             stop for irange.

            stop(int):             step for irange.

            to_skip(iter):         some form of iterable or list of elements to
                                   skip (can be a single value as well).

        Returns:
            (generator object):    an irange-like generator that skips some
                                   values.

        Examples:
            >>> xrange_with_skip(10) #doctest: +ELLIPSIS
            <generator object xrange_with_skip at 0x...>

            >>> list(xrange_with_skip(10))
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

            >>> list(xrange_with_skip(0, 10))
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

            >>> list(xrange_with_skip(1, 10))
            [1, 2, 3, 4, 5, 6, 7, 8, 9]

            >>> list(xrange_with_skip(0, 10, 2))
            [0, 2, 4, 6, 8]

            >>> list(xrange_with_skip(1, 10, 2))
            [1, 3, 5, 7, 9]

            >>> list(xrange_with_skip(10, to_skip = 2))
            [0, 1, 3, 4, 5, 6, 7, 8, 9]

            >>> list(xrange_with_skip(10, to_skip = [2, 7]))
            [0, 1, 3, 4, 5, 6, 8, 9]

            >>> list(xrange_with_skip(10, to_skip = [0]))
            [1, 2, 3, 4, 5, 6, 7, 8, 9]

            >>> list(xrange_with_skip(1, 10, to_skip = [0]))
            [1, 2, 3, 4, 5, 6, 7, 8, 9]

            >>> list(xrange_with_skip(10, to_skip = [9]))
            [0, 1, 2, 3, 4, 5, 6, 7, 8]
    """

    full = None

    if (stop is None):
        full = iter(irange(start))
    elif (step is None):
        full = iter(irange(start, stop))
    else:
        full = iter(irange(start, stop, step))

    if to_skip is None:
        to_skip = iter([])
    else:
        try:
            to_skip = iter(sorted(set(to_skip)))
        except TypeError:
            to_skip = iter([to_skip])

    next_to_skip = next(to_skip, None)

    for each in full:
        if each != next_to_skip:
            yield(each)
        else:
            next_to_skip = next(to_skip, None)


def splitting_xrange(a, b=None):
    """
        Similar to irange except that it recursively proceeds through the given
        range in such a way that values that follow each other are preferably
        not only non-sequential, but fairly different. This does not always
        work with small ranges, but works nicely with large ranges.

        Args:
            a(int):              the lower bound of the range
            b(int):              the upper bound of the range

        Returns:
            result(generator):   a generator that can be used to iterate
                                 through the sequence.

        Examples:
            >>> splitting_xrange(0, 0) #doctest: +ELLIPSIS
            <generator object splitting_xrange at 0x...>

            >>> list(splitting_xrange(0, 0))
            []

            >>> list(splitting_xrange(5, 5))
            []

            >>> list(splitting_xrange(5, 0))
            []

            >>> list(splitting_xrange(0, 1))
            [0]

            >>> list(splitting_xrange(0, 2))
            [0, 1]

            >>> list(splitting_xrange(0, 3))
            [0, 2, 1]

            >>> list(splitting_xrange(10))
            [0, 5, 8, 3, 9, 4, 6, 1, 7, 2]

            >>> list(splitting_xrange(0, 10))
            [0, 5, 8, 3, 9, 4, 6, 1, 7, 2]
    """

    def splitting_xrange_helper(a, b):
        if a != b:
            half_diff = (b - a) / 2.0

            mid_1 = int(a + math.floor(half_diff))
            mid_2 = int(a + math.ceil(half_diff))

            if a < mid_1 and b > mid_2:
                yield(mid_2)

                for _1, _2 in izip(
                        splitting_xrange_helper(a, mid_1),
                        splitting_xrange_helper(mid_2, b)
                ):
                    yield(_2)
                    yield(_1)

                if mid_1 != mid_2:
                    yield(mid_1)
            elif a < mid_1:
                for _2 in splitting_xrange_helper(mid_1, b):
                    yield(_2)
            elif b > mid_2:
                for _1 in splitting_xrange_helper(a, mid_2):
                    yield(_1)

    if b is None:
        b = a
        a = 0

    a = int(a)
    b = int(b)

    if a >= b:
        return

    yield(a)

    for each in splitting_xrange_helper(a, b):
        yield(each)


@prof.log_call(trace_logger)
def subrange(start, stop=None, step=None, substep=None):
    """
        Generates start and stop values for each subrange.

        Args:
            start(int):           First value in range (or last if only
                                  specified value)

            stop(int):            Last value in range

            step(int):            Step between each range

            substep(int):         Step within each range

        Yields:
            (irange):             A subrange within the larger range.

        Examples:
            >>> subrange(5) # doctest: +ELLIPSIS
            <generator object subrange at 0x...>

            >>> list(imap(list, subrange(5)))
            [[0], [1], [2], [3], [4]]

            >>> list(imap(list, subrange(0, 5)))
            [[0], [1], [2], [3], [4]]

            >>> list(imap(list, subrange(1, 5)))
            [[1], [2], [3], [4]]

            >>> list(imap(list, subrange(0, 10, 3)))
            [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]

            >>> list(imap(list, subrange(0, 7, 3)))
            [[0, 1, 2], [3, 4, 5], [6]]

            >>> list(imap(list, subrange(0, 7, 3, 2)))
            [[0, 2], [3, 5], [6]]
    """

    if stop is None:
        stop = start
        start = 0

    if step is None:
        step = 1

    if substep is None:
        substep = 1

    range_ends = itertools.chain(irange(start, stop, step), [stop])

    for i, j in lagged_generators_zipped(range_ends):
        yield(irange(i, j, substep))


@prof.log_call(trace_logger)
def cumulative_generator(new_op, new_iter):
    """
        Takes each value from new_iter and applies new_op to it with the result
        of previous values.

        For instance cumulative_generator(op.mul, irange(1,5)) will return all
        factorials up to and including the factorial of 4 (24).

        Args:
            new_op(callabel):      something that can be called on two values
                                   and return a result with a type that is a
                                   permissible argument.

            new_iter(iter):        an iterator or something that can be turned
                                   into an iterator.

        Returns:
            (generator object):    an iterator over the intermediate results.

        Examples:
            >>> import operator
            >>> cumulative_generator(operator.add, 10) #doctest: +ELLIPSIS
            <generator object cumulative_generator at 0x...>

            >>> list(cumulative_generator(operator.add, irange(1,5)))
            [1, 3, 6, 10]

            >>> list(cumulative_generator(operator.add, irange(5)))
            [0, 1, 3, 6, 10]

            >>> list(cumulative_generator(operator.mul, irange(5)))
            [0, 0, 0, 0, 0]

            >>> list(cumulative_generator(operator.mul, irange(1,5)))
            [1, 2, 6, 24]
    """

    new_iter = iter(new_iter)

    cur = next(new_iter)
    yield (cur)

    for each in new_iter:
        cur = new_op(cur, each)
        yield (cur)


@prof.log_call(trace_logger)
def reverse_each_element(new_iter):
    """
        Takes each element yielded by new_iter and reverses it using reversed.

        Args:
            new_iter(iter):        an iterator or something that can be turned
                                   into an iterator.

        Returns:
            (generator object):    an iterator over the reversed elements.

        Examples:
            >>> reverse_each_element(
            ...     zip(irange(5, 11), irange(5))
            ... ) #doctest: +ELLIPSIS
            <generator object reverse_each_element at 0x...>

            >>> list(reverse_each_element(zip(irange(5, 11), irange(5))))
            [(0, 5), (1, 6), (2, 7), (3, 8), (4, 9)]

            >>> list(reverse_each_element(iter([[5]])))
            [[5]]

            >>> list(reverse_each_element(iter([[5,2,3], [1, 7, 9]])))
            [[3, 2, 5], [9, 7, 1]]
    """

    new_iter = iter(new_iter)

    for each in new_iter:
        yield (type(each)(reversed(each)))


@prof.log_call(trace_logger)
def lagged_generators(new_iter, n=2):
    """
        Creates a tuple of generators with each next generator one step ahead
        of the previous generator.

        Args:
            new_iter(iter):                 an iterator or something that can
                                            be turned into an iterator

            n(int):                         number of generators to create as
                                            lagged

        Returns:
            (tuple of generator objects):   a tuple of iterators with each one
                                            step in front of the others.

        Examples:
            >>> lagged_generators(irange(5), 1) #doctest: +ELLIPSIS
            (<itertools... object at 0x...>,)

            >>> list(izip(*lagged_generators(irange(5), 1)))
            [(0,), (1,), (2,), (3,), (4,)]

            >>> list(izip(*lagged_generators(irange(5), 2)))
            [(0, 1), (1, 2), (2, 3), (3, 4)]

            >>> list(izip(*lagged_generators(irange(5))))
            [(0, 1), (1, 2), (2, 3), (3, 4)]

            >>> list(izip(*lagged_generators(irange(5), 3)))
            [(0, 1, 2), (1, 2, 3), (2, 3, 4)]

            >>> list(izip_longest(*lagged_generators(irange(5))))
            [(0, 1), (1, 2), (2, 3), (3, 4), (4, None)]

            >>> list(izip_longest(*lagged_generators(irange(5), 3)))
            [(0, 1, 2), (1, 2, 3), (2, 3, 4), (3, 4, None), (4, None, None)]
    """

    assert (n >= 0), \
        "Only a positive semi-definite number of generators can be created."

    # Where they will be stored
    all_iters = tuple()

    # If some positive definite number of generators is requested, then fill
    # the list.
    if n > 0:
        # Convert to the same type
        next_iter = itertools.tee(new_iter, 1)[0]
        for i in irange(1, n):
            # Duplicate the iterator
            prev_iter, next_iter = itertools.tee(next_iter, 2)

            # Store the copy of the old one
            all_iters += (prev_iter,)

            # Attempt to advance the next one
            # If it fails, create an empty iterator.
            try:
                next(next_iter)
            except StopIteration:
                next_iter = itertools.tee([], 1)[0]

        # Add the last one. If n == 1, the last one is the only one.
        all_iters += (next_iter,)

    return(all_iters)


@prof.log_call(trace_logger)
def lagged_generators_zipped(new_iter, n=2, longest=False, fillvalue=None):
    """
        Creates a tuple of generators with each next generator one step ahead
        of the previous generator.

        Args:
            new_iter(iter):                 an iterator or something that can
                                            be turned into an iterator

            n(int):                         number of generators to create as
                                            lagged

            longest(bool):                  whether to continue zipping along
                                            the longest generator

            fillvalue:                      value to use to fill generators
                                            shorter than the longest.

        Returns:
            generator object:               a generator object that will return
                                            values from each iterator.

        Examples:
            >>> isinstance(lagged_generators_zipped(irange(5), 1), izip)
            True

            >>> list(lagged_generators_zipped(irange(5)))
            [(0, 1), (1, 2), (2, 3), (3, 4)]

            >>> list(lagged_generators_zipped(irange(5), 1))
            [(0,), (1,), (2,), (3,), (4,)]

            >>> list(lagged_generators_zipped(irange(5), 2))
            [(0, 1), (1, 2), (2, 3), (3, 4)]

            >>> list(lagged_generators_zipped(irange(5), 3))
            [(0, 1, 2), (1, 2, 3), (2, 3, 4)]

            >>> list(lagged_generators_zipped(irange(5), longest=True))
            [(0, 1), (1, 2), (2, 3), (3, 4), (4, None)]

            >>> list(lagged_generators_zipped(irange(5), 3, True))
            [(0, 1, 2), (1, 2, 3), (2, 3, 4), (3, 4, None), (4, None, None)]
    """

    all_iters = lagged_generators(new_iter, n=n)

    zipped_iters = None
    if longest:
        zipped_iters = izip_longest(*all_iters, fillvalue=fillvalue)
    else:
        zipped_iters = izip(*all_iters)

    return(zipped_iters)


@prof.log_call(trace_logger)
def filled_stringify_numbers(new_iter, include_numbers=False):
    """
        Like enumerate except it also returns a string with the number from
        enumeration with left padding by zero.

        Args:
            new_iter(iter):        an iterator to use for enumeration over.

        Returns:
            (generator object):    an iterator over the reversed elements.

        Examples:
            >>> filled_stringify_numbers([5, 7]) #doctest: +ELLIPSIS
            <generator object filled_stringify_numbers at 0x...>

            >>> list(filled_stringify_numbers([]))
            []

            >>> list(filled_stringify_numbers(irange(5)))
            ['0', '1', '2', '3', '4']

            >>> list(filled_stringify_numbers([5]))
            ['5']

            >>> list(filled_stringify_numbers([5, 7]))
            ['5', '7']

            >>> list(filled_stringify_numbers([5, 7, 11]))
            ['05', '07', '11']

            >>> list(
            ...     filled_stringify_numbers([5, 7, 11], include_numbers=True)
            ... )
            [(5, '05'), (7, '07'), (11, '11')]

            >>> list(
            ...     filled_stringify_numbers(
            ...         iter([5, 7, 11]),
            ...         include_numbers=True
            ...     )
            ... )
            [(5, '05'), (7, '07'), (11, '11')]
    """

    new_list = new_iter
    new_list_len = None
    new_list_max = None
    try:
        new_list_len = len(new_list)
        new_list_max = max(new_list) if new_list_len else None
    except TypeError:
        new_list = list(new_list)
        new_list_len = len(new_list)
        new_list_max = max(new_list) if new_list_len else None

    if new_list_len:
        if new_list_max:
            digits = int(numpy.floor(numpy.log10(new_list_max))) + 1
        else:
            digits = 1

    if include_numbers:
        for each in new_list:
            yield((each, str(each).zfill(digits)))
    else:
        for each in new_list:
            yield(str(each).zfill(digits))


@prof.log_call(trace_logger)
def filled_stringify_xrange(new_iter):
    """
        Takes each element yielded by new_iter and reverses it using reversed.

        Args:
            new_iter(iter):        an iterator to use for enumeration over.

        Returns:
            (generator object):    an iterator over the reversed elements.

        Examples:
            >>> filled_stringify_xrange([5, 7]) #doctest: +ELLIPSIS
            <generator object filled_stringify_xrange at 0x...>

            >>> list(filled_stringify_xrange([]))
            []

            >>> list(filled_stringify_xrange(irange(5)))
            [(0, '0'), (1, '1'), (2, '2'), (3, '3'), (4, '4')]

            >>> list(filled_stringify_xrange(irange(2, 5)))
            [(0, '0'), (1, '1'), (2, '2')]

            >>> list(filled_stringify_xrange([5]))
            [(0, '0')]

            >>> list(filled_stringify_xrange([5, 7]))
            [(0, '0'), (1, '1')]

            >>> list(filled_stringify_xrange(iter([5, 7])))
            [(0, '0'), (1, '1')]

            >>> list(filled_stringify_xrange(list(irange(11))))
            [(0, '00'), (1, '01'), (2, '02'), (3, '03'), (4, '04'), (5, '05'), (6, '06'), (7, '07'), (8, '08'), (9, '09'), (10, '10')]
    """

    new_list = new_iter
    new_list_index_gen = None
    try:
        new_list_index_gen = irange(len(new_list))
    except TypeError:
        new_list = list(new_list)
        new_list_index_gen = irange(len(new_list))

    new_list_index_gen_stringified = filled_stringify_numbers(
        new_list_index_gen, include_numbers=True
    )

    for i, i_str in new_list_index_gen_stringified:
        yield ((i, i_str))


@prof.log_call(trace_logger)
def filled_stringify_enumerate(new_iter):
    """
        Takes each element yielded by new_iter and reverses it using reversed.

        Args:
            new_iter(iter):        an iterator to use for enumeration over.

        Returns:
            (generator object):    an iterator over the reversed elements.

        Examples:
            >>> filled_stringify_enumerate([5, 7]) #doctest: +ELLIPSIS
            <generator object filled_stringify_enumerate at 0x...>

            >>> list(filled_stringify_enumerate([]))
            []

            >>> list(filled_stringify_enumerate(irange(5)))
            [(0, '0', 0), (1, '1', 1), (2, '2', 2), (3, '3', 3), (4, '4', 4)]

            >>> list(filled_stringify_enumerate(irange(2, 5)))
            [(0, '0', 2), (1, '1', 3), (2, '2', 4)]

            >>> list(filled_stringify_enumerate([5]))
            [(0, '0', 5)]

            >>> list(filled_stringify_enumerate([5, 7]))
            [(0, '0', 5), (1, '1', 7)]

            >>> list(filled_stringify_enumerate(iter([5, 7])))
            [(0, '0', 5), (1, '1', 7)]

            >>> list(filled_stringify_enumerate(range(11)))
            [(0, '00', 0), (1, '01', 1), (2, '02', 2), (3, '03', 3), (4, '04', 4), (5, '05', 5), (6, '06', 6), (7, '07', 7), (8, '08', 8), (9, '09', 9), (10, '10', 10)]
    """

    new_list = new_iter
    new_list_index_gen = None
    try:
        new_list_index_gen = irange(len(new_list))
    except TypeError:
        new_list = list(new_list)
        new_list_index_gen = irange(len(new_list))

    new_list_index_gen_stringified = filled_stringify_numbers(
        new_list_index_gen, include_numbers=True
    )

    for (i, i_str), each in izip(new_list_index_gen_stringified, new_list):
        yield ((i, i_str, each))


@prof.log_call(trace_logger)
def reformat_slice(a_slice, a_length=None):
    """
        Takes a slice and reformats it to fill in as many undefined values as
        possible.

        Args:
            a_slice(slice):        a slice to reformat.

            a_length(int):         a length to fill for stopping if not
                                   provided.

        Returns:
            (slice):               a new slice with as many values filled in as
                                   possible.

        Examples:
            >>> reformat_slice(slice(None))
            slice(0, None, 1)

            >>> reformat_slice(slice(None), 10)
            slice(0, 10, 1)

            >>> reformat_slice(slice(2, None))
            slice(2, None, 1)

            >>> reformat_slice(slice(2, None), 10)
            slice(2, 10, 1)

            >>> reformat_slice(slice(2, None, None))
            slice(2, None, 1)

            >>> reformat_slice(slice(2, None, None), 10)
            slice(2, 10, 1)

            >>> reformat_slice(slice(2, -1, None), 10)
            slice(2, 9, 1)

            >>> range(10)[reformat_slice(slice(None))] == range(10)[:]
            True

            >>> range(10)[reformat_slice(slice(2, None))] == range(10)[2:]
            True

            >>> range(10)[reformat_slice(slice(2, 6))] == range(10)[2:6]
            True

            >>> range(10)[reformat_slice(slice(2, 6, 3))] == range(10)[2:6:3]
            True

            >>> range(10)[reformat_slice(slice(2, None, 3))] == range(10)[2::3]
            True

            >>> range(10)[reformat_slice(slice(None), 10)] == range(10)[:]
            True

            >>> range(10)[reformat_slice(slice(2, None), 10)] == range(10)[2:]
            True

            >>> range(10)[reformat_slice(slice(2, 6), 10)] == range(10)[2:6]
            True

            >>> range(10)[reformat_slice(slice(2, 6, 3), 10)] == range(10)[2:6:3]
            True

            >>> range(10)[reformat_slice(slice(2, None, 3), 10)] == range(10)[2::3]
            True

            >>> range(10)[reformat_slice(slice(2, -6, 3), 10)] == range(10)[2:-6:3]
            True

            >>> range(10)[reformat_slice(slice(2, -1), 10)] == range(10)[2:-1]
            True

            >>> range(10)[reformat_slice(slice(2, 20), 10)] == range(10)[2:20]
            True

            >>> range(10)[reformat_slice(slice(2, -20), 10)] == range(10)[2:-20]
            True

            >>> range(10)[reformat_slice(slice(20, -1), 10)] == range(10)[20:-1]
            True

            >>> range(10)[reformat_slice(slice(-20, -1), 10)] == range(10)[-20:-1]
            True

            >>> range(10)[reformat_slice(slice(-5, -1), 10)] == range(10)[-5:-1]
            True
    """

    assert (a_slice is not None), "err"

    # Fill unknown values.

    new_slice_step = a_slice.step
    if new_slice_step is None:
        new_slice_step = 1

    new_slice_start = a_slice.start
    if (new_slice_start is None) and (new_slice_step > 0):
        new_slice_start = 0

    new_slice_stop = a_slice.stop


    # Make adjustments for length

    if a_length is not None:
        if (new_slice_step < -a_length):
            new_slice_step = -a_length
        elif (new_slice_step > a_length):
            new_slice_step = a_length

        if (new_slice_start is None) and (new_slice_step > 0):
            pass
        elif (new_slice_start is None) and (new_slice_step < 0):
            new_slice_start = a_length
        elif (new_slice_start <= -a_length) and (new_slice_step > 0):
            new_slice_start = 0
        elif (new_slice_start < -a_length) and (new_slice_step < 0):
            new_slice_start = new_slice_stop = 0
        elif (new_slice_start > a_length) and (new_slice_step > 0):
            new_slice_start = a_length
        elif (new_slice_start > a_length) and (new_slice_step < 0):
            new_slice_start = a_length
        elif (new_slice_start < 0) and (new_slice_step > 0):
            new_slice_start += a_length
        elif (new_slice_start < 0) and (new_slice_step < 0):
            new_slice_start += a_length

        if (new_slice_stop is None) and (new_slice_step > 0):
            new_slice_stop = a_length
        elif (new_slice_stop is None) and (new_slice_step < 0):
            pass
        elif (new_slice_stop <= -a_length) and (new_slice_step > 0):
            new_slice_stop = 0
        elif (new_slice_stop < -a_length) and (new_slice_step < 0):
            new_slice_stop = None
        elif (new_slice_stop < 0) and (new_slice_step > 0):
            new_slice_stop += a_length
        elif (new_slice_stop < 0) and (new_slice_step < 0):
            new_slice_stop += a_length
        elif (new_slice_stop > a_length) and (new_slice_step > 0):
            new_slice_stop = a_length
        elif (new_slice_stop >= a_length) and (new_slice_step < 0):
            new_slice_start = new_slice_stop = 0
        elif (new_slice_stop < 0) and (new_slice_step > 0):
            new_slice_stop += a_length
        elif (new_slice_stop < 0) and (new_slice_step < 0):
            new_slice_stop += a_length


    # Build new slice and return.

    new_slice = slice(new_slice_start, new_slice_stop, new_slice_step)

    return(new_slice)


@prof.log_call(trace_logger)
def reformat_slices(slices, lengths=None):
    """
        Takes a tuple of slices and reformats them to fill in as many undefined
        values as possible.

        Args:
            slices(tuple(slice)):        a tuple of slices to reformat.
            lengths(tuple(int)):         a tuple of lengths to fill.

        Returns:
            (slice):                     a tuple of slices with all default
                                         values filled if possible.

        Examples:

            >>> reformat_slices(slice(None))
            (slice(0, None, 1),)

            >>> reformat_slices((slice(None),))
            (slice(0, None, 1),)

            >>> reformat_slices((
            ...     slice(None),
            ...     slice(3, None),
            ...     slice(None, 5),
            ...     slice(None, None, 2)
            ... ))
            (slice(0, None, 1), slice(3, None, 1), slice(0, 5, 1), slice(0, None, 2))

            >>> reformat_slices(
            ...     (
            ...         slice(None),
            ...         slice(3, None),
            ...         slice(None, 5),
            ...         slice(None, None, 2)
            ...     ),
            ...     (10, 13, 15, 20)
            ... )
            (slice(0, 10, 1), slice(3, 13, 1), slice(0, 5, 1), slice(0, 20, 2))
    """

    try:
        len(slices)
    except TypeError:
        slices = (slices,)

    new_lengths = lengths
    if new_lengths is None:
        new_lengths = [None] * len(slices)

    try:
        len(new_lengths)
    except TypeError:
        new_lengths = (new_lengths,)

    assert (len(slices) == len(new_lengths))

    new_slices = list(slices)
    for i, each_length in enumerate(new_lengths):
        new_slices[i] = reformat_slice(new_slices[i], each_length)

    new_slices = tuple(new_slices)

    return(new_slices)


class UnknownSliceLengthException(Exception):
    """
        Raised if a slice does not have a known length.
    """

    pass


@prof.log_call(trace_logger)
def len_slice(a_slice, a_length=None):
    """
        Determines how many elements a slice will contain.

        Raises:
            UnknownSliceLengthException: Will raise an exception if
            a_slice.stop and a_length is None.

        Args:
            a_slice(slice):        a slice to reformat.

            a_length(int):         a length to fill for stopping if not
                                   provided.

        Returns:
            (slice):               a new slice with as many values filled in as
                                   possible.

        Examples:
            >>> len_slice(slice(None)) #doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            UnknownSliceLengthException: Cannot determine slice length without a defined end point. The reformatted slice was slice(0, None, 1).

            >>> len_slice(slice(None), 10)
            10

            >>> len_slice(slice(None), 10) == len(range(10)[:])
            True

            >>> len_slice(slice(2, None), 10)
            8

            >>> len_slice(slice(2, None), 10) == len(range(10)[2:])
            True

            >>> len_slice(slice(2, None, None), 10)
            8

            >>> len_slice(slice(2, None, None), 10) == len(range(10)[2:])
            True

            >>> len_slice(slice(2, 6))
            4

            >>> len_slice(slice(2, 6), 1000)
            4

            >>> len_slice(slice(2, 6), 10) == len(range(10)[2:6])
            True

            >>> len_slice(slice(2, 6, 3))
            2

            >>> len_slice(slice(2, 6, 3), 10) == len(range(10)[2:6:3])
            True
    """

    new_slice = reformat_slice(a_slice, a_length)

    if new_slice.stop is None:
        raise UnknownSliceLengthException(
            "Cannot determine slice length without a defined end point. " +
            "The reformatted slice was " + repr(new_slice) + "."
        )

    new_slice_diff = float(new_slice.stop - new_slice.start)

    new_slice_size = int(math.ceil(new_slice_diff / new_slice.step))

    return(new_slice_size)


@prof.log_call(trace_logger)
def len_slices(slices, lengths=None):
    """
        Takes a tuple of slices and reformats them to fill in as many undefined
        values as possible.

        Args:
            slices(tuple(slice)):        a tuple of slices to reformat.
            lengths(tuple(int)):         a tuple of lengths to fill.

        Returns:
            (slice):                     a tuple of slices with all default
                                         values filled if possible.

        Examples:
            >>> len_slices((
            ...     slice(None),
            ...     slice(3, None),
            ...     slice(None, 5),
            ...     slice(None, None, 2)
            ... )) #doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            UnknownSliceLengthException: Cannot determine slice length without a defined end point. The reformatted slice was slice(0, None, 1).

            >>> len_slices(
            ...     (
            ...         slice(None),
            ...         slice(3, None),
            ...         slice(None, 5),
            ...         slice(None, None, 2)
            ...     ),
            ...     (10, 13, 15, 20)
            ... )
            (10, 10, 5, 10)
    """

    new_slices = reformat_slices(slices, lengths)

    lens = []

    for each_slice in new_slices:
        lens.append(len_slice(each_slice))

    lens = tuple(lens)

    return(lens)



###############################################################################
#   lazyflow: data flow based lazy parallel computation framework
#
#       Copyright (C) 2011-2014, the ilastik developers
#                                <team@ilastik.org>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the Lesser GNU General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# See the files LICENSE.lgpl2 and LICENSE.lgpl3 for full text of the
# GNU Lesser General Public License version 2.1 and 3 respectively.
# This information is also available on the ilastik web site at:
#                 http://ilastik.org/license/
###############################################################################
import os


class PathComponents(object):
    """
    Provides a convenient access to path components of a combined external/internal path to a dataset.
    Also, each of the properties listed below is writable, in which case ALL properties are updated accordingly.
    """

    # Only files with these extensions are allowed to have an 'internal' path
    HDF5_EXTS = ['.ilp', '.h5', '.hdf5']

    def __init__(self, totalPath, cwd=None):
        """
        Initialize the path components.

        :param totalPath: The entire path to the dataset, including any internal path (e.g. the path to an hdf5 dataset).
                          For example, ``totalPath='/some/path/to/file.h5/with/internal/dataset'``

        :param cwd: If provided, relative paths will be converted to absolute paths using this arg as the working directory.
        """
        self._externalPath = None
        self._externalDirectory = None
        self._filename = None
        self._filenameBase = None
        self._extension = None
        self._internalPath = None
        self._internalDatasetName = None
        self._internalDirectory = None

        self._cwd = cwd
        self._init(totalPath, cwd)
        self._initialized = True

    def _init(self, totalPath, cwd):
        ext = None
        extIndex = -1

        if cwd is not None:
            absPath, relPath = getPathVariants(totalPath, cwd)
            totalPath = absPath

        # convention for Windows: use "/"
        totalPath = totalPath.replace("\\","/")

        # For hdf5 paths, split into external, extension, and internal paths
        for x in self.HDF5_EXTS:
            if totalPath.find(x) > extIndex:
                extIndex = totalPath.find(x)
                ext = x

        # Comments below refer to this example path:
        # /some/path/to/file.h5/with/internal/dataset
        if ext is not None:
            self._extension = ext                              # .h5
            parts = totalPath.split(ext)

            # Must deal with pathological filenames such as /path/to/file.h5_with_duplicate_ext.h5
            while len(parts) > 2:
                parts[0] = parts[0] + ext + parts[1]
                del parts[1]
            self._externalPath = parts[0] + ext # /some/path/to/file.h5
            self._internalPath = parts[1].replace('\\', '/') # /with/internal/dataset

            self._internalDirectory = os.path.split(self.internalPath)[0]   # /with/internal
            self._internalDatasetName = os.path.split(self.internalPath)[1] # dataset
        else:
            # For non-hdf5 files, use normal path/extension (no internal path)
            (self._externalPath, self._extension) = os.path.splitext(totalPath)
            self._externalPath += self._extension
            self._internalPath = None
            self._internalDatasetName = None
            self._internalDirectory = None

        self._externalDirectory = os.path.split(self._externalPath)[0] # /some/path/to
        self._filename = os.path.split(self._externalPath)[1]          # file.h5
        self._filenameBase = os.path.splitext(self._filename)[0]         # file

    def __setattr__(self, attr, value):
        """
        This prevents us from accidentally writing to a non-existant attribute.
        e.g. if the user tries to say components.fIlEnAmE = 'newfile.h5', raise an error.
        """
        if hasattr(self, '_initialized'):
            # Attempt to get the old attribute first
            oldval = getattr(self, attr)
        object.__setattr__(self, attr, value)

    def totalPath(self):
        """
        Return the (reconstructed) totalPath to the dataset.
        """
        total = self.externalPath
        if self.internalPath:
            total += self.internalPath
        return total

    #
    # Getters
    #

    @property
    def externalPath(self):
        """
        Example: ``/some/path/to/file.h5``
        """
        return self._externalPath

    @property
    def externalDirectory(self):
        """
        Example: ``/some/path/to``
        """
        return self._externalDirectory

    @property
    def filename(self):
        """
        Example: ``file.h5``
        """
        return self._filename

    @property
    def filenameBase(self):
        """
        Example: ``file``
        """
        return self._filenameBase

    @property
    def extension(self):
        """
        Example: ``.h5``
        """
        return self._extension

    @property
    def internalPath(self):
        """
        Example: ``/with/internal/dataset``
        """
        return self._internalPath

    @property
    def internalDatasetName(self):
        """
        Example: ``/dataset``
        """
        return self._internalDatasetName

    @property
    def internalDirectory(self):
        """
        Example: ``/with/internal``
        """
        return self._internalDirectory

    #
    # Setters
    #

    @externalPath.setter
    def externalPath(self, new):
        assert new[-1] != '/'
        if self._internalPath:
            self._init(new + self._internalPath, self._cwd)
        else:
            self._init(new, self._cwd)

    @externalDirectory.setter
    def externalDirectory(self, new):
        new_external = os.path.join(new, self._filename)
        self.externalPath = new_external

    @filename.setter
    def filename(self, new):
        new_external = os.path.join(self._externalDirectory, new)
        self.externalPath = new_external

    @filenameBase.setter
    def filenameBase(self, new):
        new_external = os.path.join(self._externalDirectory, new + self._extension)
        self.externalPath = new_external

    @extension.setter
    def extension(self, new):
        assert (not self._internalPath) or (new in self.HDF5_EXTS), \
            "This PathComponents has an internal path ({}), but you are "\
            "attempting to assign a non-hdf5 extension to it ({})"\
            .format(self._internalPath, new)
        new_external = os.path.join(self._externalDirectory, self._filenameBase + new)
        self.externalPath = new_external

    @internalPath.setter
    def internalPath(self, new):
        assert self._extension in self.HDF5_EXTS, \
            "Can't set an internal path on a filename with extension {}".format(self._extension)
        if new:
            self._init(self._externalPath + new, self._cwd)
        else:
            self._init(self._externalPath, self._cwd)

    @internalDatasetName.setter
    def internalDatasetName(self, new):
        new_internal = os.path.join(self._internalDirectory, new)
        self.internalPath = new_internal

    @internalDirectory.setter
    def internalDirectory(self, new):
        if new and new[0] != '/':
            new = '/' + new
        new_internal = os.path.join(new, self._internalDatasetName)
        self.internalPath = new_internal


def areOnSameDrive(path1,path2):
    # if one path is relative, assume they are on same drive
    if isUrl(path1) or isUrl(path2):
        return False
    if not os.path.isabs(path1) or not os.path.isabs(path2):
        return True
    drive1,path1 = os.path.splitdrive(path1)
    drive2,path2 = os.path.splitdrive(path2)
    return drive1==drive2


def compressPathForDisplay(pathstr,maxlength):
    '''Add alternatingly parts of the start and the end of the path
    until maxlength is exceeded. Result: Drive/Dir1/.../Dirn/file'''
    if len(pathstr) <= maxlength:
        return pathstr
    dots = '...'
    component_list = pathstr.split("/")
    prefix = ""
    suffix = "/" + component_list.pop(-1)
    while component_list:
        restlength = maxlength - len(prefix) - len(dots)
        suffix = '/' + component_list.pop(-1) + suffix
        if len(suffix) > restlength:
            suffix = suffix[-restlength:]
            break
        if not component_list:
            break
        c = prefix + component_list.pop(0) + "/"
        if len(c)+len(dots)+len(suffix) > maxlength:
            break
        prefix = c
    return prefix + dots + suffix


def isUrl(path):
    # For now, the simplest rule will work.
    return '://' in path


def getPathVariants(originalPath, workingDirectory):
    """
    Take the given filePath (which can be absolute or relative, and may include an internal path suffix),
    and return a tuple of the absolute and relative paths to the file.
    """
    # urls are considered absolute
    if isUrl(originalPath):
        return originalPath, None

    if len(originalPath) > 0 and originalPath[0] == '~':
        originalPath = os.path.expanduser(originalPath)

    relPath = originalPath

    if os.path.isabs(originalPath):
        absPath = originalPath
        if areOnSameDrive(originalPath,workingDirectory):
            relPath = os.path.relpath(absPath, workingDirectory).replace("\\","/")
        else:
            # Relative path does not always exist.  Caller must check for None.
            relPath = None
    else:
        relPath = originalPath
        absPath = os.path.normpath(os.path.join(workingDirectory, relPath))

    return (absPath, relPath)



"""
The module ``serializers`` performs IO of NumPy object to an from HDF5 files.

===============================================================================
Overview
===============================================================================
The module ``serializers`` provides an easy way to serialize unusual NumPy
object to an from files. In particular, it provides support for structured
arrays and masked arrays.

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$May 14, 2014 22:19:59 EDT$"


import os

import numpy
import h5py

from nanshe.util.pathHelpers import PathComponents
from nanshe.util import wrappers

# Need in order to have logging information no matter what.
from nanshe.util import prof


try:
    unicode
except NameError:
    unicode = str


# Get the logger
trace_logger = prof.getTraceLogger(__name__)


@prof.log_call(trace_logger)
def create_numpy_structured_array_in_HDF5(file_handle,
                                          internalPath,
                                          data,
                                          overwrite=False):
    """
        Serializes a NumPy structure array to an HDF5 file by using the HDF5
        compound data type. Also, will handle normal NumPy arrays and scalars,
        as well.

        Note:
            HDF5 does not support generic Python objects. So, serialization of
            objects to something else (perhaps strs of fixed size) must be
            performed first.

        Args:
            file_handle(HDF5 file):     either an HDF5 filename or Group.
            internalPath(str):          an internal path for the HDF5 file.
            data(numpy.ndarray):        the NumPy structure array to save (or
                                        normal NumPy array).

            overwrite(bool):            whether to overwrite what is already
                                        there (defaults to False).
    """

    close_file_handle = False

    if isinstance(file_handle, str) or isinstance(file_handle, unicode):
        file_handle = h5py.File(file_handle, "a")
        close_file_handle = True

    data_array = data
    if not isinstance(data_array, numpy.ndarray):
        try:
            data_array = numpy.array(data_array)
        except:
            if not data_array.dtype.names:
                raise TypeError(
                    "The argument provided for data is type: \"" +
                    repr(type(data)) + "\" is not convertible to type \"" +
                    repr(numpy.ndarray) + "\"."
                )


    try:
        file_handle.create_dataset(
            internalPath,
            shape=data_array.shape,
            dtype=data_array.dtype,
            data=data_array,
            chunks=bool(data_array.ndim)
        )
    except RuntimeError:
        if overwrite:
            del file_handle[internalPath]
            file_handle.create_dataset(
                internalPath,
                shape=data_array.shape,
                dtype=data_array.dtype,
                data=data_array,
                chunks=True
            )
        else:
            raise

    if close_file_handle:
        file_handle.close()


@prof.log_call(trace_logger)
def read_numpy_structured_array_from_HDF5(file_handle, internalPath):
    """
        Serializes a NumPy structure array from an HDF5 file by using the HDF5
        compound data type. Also, it will handle normal NumPy arrays and
        scalars, as well.

        Note:
            HDF5 does not support generic Python objects. So, serialization of
            objects to something else (perhaps strs of fixed size) must be
            performed first.

        Args:
            file_handle(HDF5 file):     either an HDF5 filename or Group.
            internalPath(str):          an internal path for the HDF5 file.

        Note:
            TODO: Write doctests.

        Returns:
            data(numpy.ndarray):    the NumPy structure array.
    """

    close_file_handle = False

    if isinstance(file_handle, str) or isinstance(file_handle, unicode):
        file_handle = h5py.File(file_handle, "r")
        close_file_handle = True

    data = None

    data_object = file_handle[internalPath]
    data_file = data_object.file
    data_ref = data_object.value
    # data_ref = data_object[()]

    if isinstance(data_ref, h5py.Reference):
        if ("filename" in data_object.attrs) and \
           (os.path.normpath(data_object.attrs["filename"]) != os.path.normpath(data_file.filename)):
            with h5py.File(data_object.attrs["filename"], "r") as external_file_handle:
                if isinstance(data_ref, h5py.RegionReference):
                    data = external_file_handle[data_ref][data_ref]
                else:
                    data = external_file_handle[data_ref].value
        else:
            if isinstance(data_ref, h5py.RegionReference):
                data = data_file[data_ref][data_ref]
            else:
                data = data_file[data_ref].value
    elif isinstance(data_ref, (numpy.number, numpy.ndarray,)):
        if ("filename" in data_object.attrs):
            # It's a pseudo-ref.
            assert ("dataset" in data_object.attrs)

            new_dataset_name = data_object.attrs["dataset"]
            with h5py.File(data_object.attrs["filename"], "r") as external_file_handle:
                # assert isinstance(new_dataset_name, h5py.Dataset)

                if ("field" in data_object.attrs) and \
                        ("slice" in data_object.attrs):
                    new_field = data_object.attrs["field"]
                    new_slicing = eval(data_object.attrs["slice"])

                    data = external_file_handle[
                        new_dataset_name][new_field][new_slicing]
                elif ("field" in data_object.attrs):
                    new_field = data_object.attrs["field"]

                    data = external_file_handle[new_dataset_name][new_field]
                elif ("slice" in data_object.attrs):
                    new_slicing = eval(data_object.attrs["slice"])

                    data = external_file_handle[new_dataset_name][new_slicing]
                else:
                    data = external_file_handle[new_dataset_name][()]
        else:
            data = data_object.value

    if close_file_handle:
        file_handle.close()

    return(data)


def hdf5_wrapper(hdf5_args=[], hdf5_kwargs=[], hdf5_result=""):
    """
        Drop array results into HDF5 files specified.

        Useful wrapper, which take a callable and handle its input arguments
        that are HDF5 Datasets and reads them in as NumPy arrays. These NumPy
        arrays are then provided to the decorated callable as normal arguments.
        The result is then stored as an HDF5 Dataset.

        Args:
            hdf5_args(Sequence):     A sequence of indices that represent
                                     arguments passed in that are expected
                                     to be HDF5 Datasets that will be read in
                                     and provided as NumPy arrays.

            hdf5_kwargs(Sequence):   A sequence of keyword arguments that are
                                     expected to be HDF5 Datasets that will be
                                     read in and provided as NumPy arrays.

            hdf5_result(bytes):      Which HDF5 Dataset to use for storing the
                                     result.

        Returns:
            callable:                Does the actual decoration.
    """

    def hdf5_decorator(a_callable):
        """
            Decorates the callable and returns the result.

            Args:
                a_callable(callable):   A callable to be wrapped to offload
                                        some arguments to HDF5 files.

            Returns:
                callable:               The decorated function with a different
                                        argument spec.
        """

        @wrappers.wraps(a_callable)
        @wrappers.static_variables(hdf5_args=hdf5_args,
                                   hdf5_kwargs=hdf5_kwargs,
                                   hdf5_result=hdf5_result)
        def hdf5_wrapped(*args, **kwargs):
            """
                Replaces the decorated callable.

                Args:
                    *args(Sequence):     Arguments for the callable.
                    **kwargs(Mapping):   Keyword arguments for the callable.

                Returns:
                    str:                 Path to where the Dataset was stored.
            """

            new_args = []
            for i, each_arg in enumerate(args):
                each_new_arg = each_arg

                if i in hdf5_wrapped.hdf5_args:
                    each_arg_pc = PathComponents(each_arg)
                    each_filename, each_datasetpath = each_arg_pc.externalPath, \
                                                      each_arg_pc.internalPath
                    with h5py.File(each_filename, "r") as each_file:
                        each_new_arg = each_file[each_datasetpath][...]

                new_args.append(each_new_arg)

            new_kwargs = dict()
            for each_key, each_kwarg in kwargs.items():
                each_new_kwarg = each_kwarg

                if each_key in hdf5_wrapped.hdf5_kwargs:
                    each_kwarg_pc = PathComponents(each_kwarg)
                    each_filename, each_datasetpath = each_kwarg_pc.externalPath, \
                                                      each_kwarg_pc.internalPath
                    with h5py.File(each_filename, "r") as each_file:
                        each_new_kwarg = each_file[each_datasetpath][...]

                new_kwargs[each_key] = each_new_kwarg

            result = a_callable(*new_args, **new_kwargs)

            if hdf5_wrapped.hdf5_result:
                result_pc = PathComponents(hdf5_wrapped.hdf5_result)
                result_filename, result_datasetpath = result_pc.externalPath, \
                                                      result_pc.internalPath
                with h5py.File(result_filename, "a") as result_file:
                    result_file[result_datasetpath] = result

                result = hdf5_wrapped.hdf5_result

            return(result)

        return(hdf5_wrapped)

    return(hdf5_decorator)


class HDF5MaskedDataset(object):
    """
        Provides an abstraction of the masked array the HDF5 Group where the
        contents of a masked array are serialized.

        Note:
                This behaves roughly like an `h5py.Dataset` and roughly like a
                `numpy.ma.masked_array`. Internally, it uses an `h5py.Group` to
                contain the components of the masked array and allow
                interaction with them.
    """

    def __init__(self,
                 group, shape=None, dtype=None, data=None, chunks=True,
                 **kwargs):
        assert isinstance(group, h5py.Group)

        assert "compression" not in kwargs
        assert "compression_opts" not in kwargs

        self._group = group

        if len(self._group):
            assert len(self._group) == 3

            assert data is None
            assert shape is None
            assert dtype is None

            assert "data" in self._group
            assert "mask" in self._group
            assert "fill_value" in self._group

            assert self._group["data"].shape == self._group["mask"].shape
            assert self._group["data"].dtype == self._group["fill_value"].dtype
            assert self._group["mask"].dtype == numpy.dtype(numpy.bool8)
        else:
            assert (data is not None) or \
                   ((shape is not None) and (dtype is not None))

            shape = tuple(shape)
            dtype = numpy.dtype(dtype)

            if data is not None:
                if (shape is not None) and (dtype is not None):
                    assert data.shape == shape
                    assert data.dtype == dtype
                else:
                    shape = tuple(data.shape)
                    dtype = numpy.dtype(data.dtype)

            self._group.create_dataset(
                "data",
                shape=shape,
                dtype=dtype,
                chunks=chunks,
                **kwargs
            )
            self._group.create_dataset(
                "mask",
                shape=shape,
                dtype=numpy.bool8,
                chunks=chunks,
                compression="gzip",
                compression_opts=2,
                **kwargs
            )
            self._group.create_dataset(
                "fill_value",
                shape=tuple(),
                dtype=dtype
            )

            if data is not None:
                self._group["data"][...] = data
                self._group["mask"][...] = numpy.ma.getmaskarray(data)
                if isinstance(data, numpy.ma.masked_array):
                    self._group["fill_value"][...] = dtype.type(
                        data.fill_value)

    @property
    def group(self):
        return(self._group)

    @property
    def name(self):
        return(self._group.name)

    @property
    def data(self):
        return(self._group["data"])

    @data.setter
    def data(self, value):
        self._group["data"][...] = value

    @property
    def mask(self):
        return(self._group["mask"])

    @mask.setter
    def mask(self, value):
        self._group["mask"][...] = value

    @property
    def fill_value(self):
        return(self._group["fill_value"])

    @fill_value.setter
    def fill_value(self, value):
        self._group["fill_value"][...] = value

    def __len__(self):
        return(len(self._group["data"]))

    @property
    def dims(self):
        return(self._group["data"].dims)

    @property
    def ndim(self):
        return(len(self._group["data"].shape))

    @property
    def shape(self):
        return(self._group["data"].shape)

    @shape.setter
    def shape(self, shape):
        self.resize(shape)

    @property
    def size(self):
        return(self._group["data"].size)

    @property
    def dtype(self):
        return(self._group["data"].dtype)

    def resize(self, size, axis=None):
        self._group["data"].resize(size, axis)
        self._group["mask"].resize(size, axis)

    def __getitem__(self, args):
        result = self._group["data"][args]
        result = result.view(numpy.ma.masked_array)

        result.mask = self._group["mask"][args]
        result.fill_value = self._group["fill_value"][...]

        return(result)

    def __setitem__(self, args, value):
        self._group["data"][args] = value
        self._group["mask"][args] = numpy.ma.getmaskarray(value)

        if isinstance(value, numpy.ma.masked_array):
            self._group["fill_value"][...] = value.fill_value

    def __array__(self, dtype=None):
        result = self._group["data"].__array__(dtype=dtype)
        result = result.view(numpy.ma.masked_array)

        result.mask = self._group["mask"][...]
        result.fill_value = self._group["fill_value"][...]

        return(result)



"""
The ``registerer`` module allows the registration algorithm to be run.

===============================================================================
Overview
===============================================================================
The ``main`` function actually starts the algorithm and can be called
externally. Configuration files for the registerer are provided in the
examples_ and are entitled registerer. Any attributes on the raw dataset are
copied to the registered dataset.

.. _examples: http://github.com/nanshe-org/nanshe/tree/master/examples

===============================================================================
API
===============================================================================
"""


__author__ = "John Kirkham <kirkhamj@janelia.hhmi.org>"
__date__ = "$Feb 20, 2015 13:00:51 EST$"


import itertools
import os

import h5py

from nanshe.util import iters, prof
from nanshe.io import xjson
from nanshe.util.pathHelpers import PathComponents
from nanshe.imp import registration



# Get the logger
trace_logger = prof.getTraceLogger(__name__)



@prof.log_call(trace_logger)
def main(*argv):
    """
        Simple main function (like in C). Takes all arguments (as from
        sys.argv) and returns an exit status.

        Args:
            argv(list):     arguments (includes command line call).

        Returns:
            int:            exit code (0 if success)
    """

    # Only necessary if running main (normally if calling command line). No
    # point in importing otherwise.
    import argparse

    argv = list(argv)

    # Creates command line parser
    parser = argparse.ArgumentParser(
        description="Parses input from the command line " +
                    "for a registration job."
    )

    parser.add_argument("config_filename",
                        metavar="CONFIG_FILE",
                        type=str,
                        help="JSON file that provides configuration options " +
                             "for how to import TIFF(s)."
    )
    parser.add_argument("input_filenames",
                        metavar="INPUT_FILE",
                        type=str,
                        nargs=1,
                        help="HDF5 file to import (this should include a " +
                             "path to where the internal dataset should be " +
                             "stored)."
    )

    parser.add_argument("output_filenames",
                        metavar="OUTPUT_FILE",
                        type=str,
                        nargs=1,
                        help="HDF5 file to export (this should include a " +
                             "path to where the internal dataset should be " +
                             "stored)."
    )

    # Results of parsing arguments
    # (ignore the first one as it is the command line call).
    parsed_args = parser.parse_args(argv[1:])

    # Go ahead and stuff in parameters with the other parsed_args
    parsed_args.parameters = xjson.read_parameters(parsed_args.config_filename)

    parsed_args.input_file_components = []
    for each_input_filename in parsed_args.input_filenames:
        parsed_args.input_file_components.append(
            PathComponents(each_input_filename)
        )

    parsed_args.output_file_components = []
    for each_output_filename in parsed_args.output_filenames:
        parsed_args.output_file_components.append(
            PathComponents(each_output_filename)
        )

    for each_input_filename_components, each_output_filename_components in iters.izip(
            parsed_args.input_file_components, parsed_args.output_file_components):
        with h5py.File(each_input_filename_components.externalPath, "r") as input_file:
            with h5py.File(each_output_filename_components.externalPath, "a") as output_file:
                data = input_file[each_input_filename_components.internalPath]
                result_filename = registration.register_mean_offsets(
                    data, to_truncate=True, **parsed_args.parameters
                )
                with h5py.File(result_filename, "r") as result_file:
                    result_file.copy(
                        "reg_frames",
                        output_file[each_output_filename_components.internalDirectory],
                        name=each_output_filename_components.internalDatasetName
                    )

                    if parsed_args.parameters.get("include_shift", False):
                        result_file.copy(
                            "space_shift",
                            output_file[each_output_filename_components.internalDirectory],
                            name=each_output_filename_components.internalDatasetName + "_shift"
                        )

                # Copy all attributes from raw data to the final result.
                output = output_file[
                    each_output_filename_components.internalDatasetName
                ]
                for each_attr_name in data.attrs:
                    output.attrs[each_attr_name] = data.attrs[each_attr_name]

                # Only remove the directory if our input or output files are
                # not stored there.
                os.remove(result_filename)
                in_out_dirnames = set(
                    os.path.dirname(os.path.abspath(_.filename)) for _ in [
                        input_file, output_file
                    ]
                )
                result_dirname = os.path.dirname(result_filename)
                if result_dirname not in in_out_dirnames:
                    os.rmdir(result_dirname)

    return(0)
